import logging
import warnings
from typing import *
from datetime import datetime, timedelta, timezone

from pydantic import model_validator, PrivateAttr

from config_wrangler.config_templates.password_source import PasswordSource

try:
    # noinspection PyPackageRequirements
    from sqlalchemy.engine import Engine
    # noinspection PyPackageRequirements
    from sqlalchemy.sql.schema import DEFAULT_NAMING_CONVENTION
    # noinspection PyPackageRequirements
    from sqlalchemy import create_engine, event, MetaData
    # noinspection PyPackageRequirements
    from sqlalchemy.engine.url import URL
    # noinspection PyPackageRequirements
    from sqlalchemy.orm import Session
    # noinspection PyPackageRequirements
    from sqlalchemy.pool import QueuePool, NullPool
    # noinspection PyPackageRequirements
    from sqlalchemy.exc import SQLAlchemyError
except ImportError:
    raise ImportError("SQLAlchemyDatabase requires sqlalchemy to be installed")

from config_wrangler.config_templates.credentials import Credentials

if TYPE_CHECKING:
    from mypy_boto3_redshift.client import RedshiftClient


class SQLAlchemyDatabase(Credentials):
    dialect: str
    """
    The SQLAlchemy dialect to use.  See https://docs.sqlalchemy.org/en/20/dialects/
    """

    driver: Optional[str] = None
    """
    The python database driver to use. 
    This is optional in the SQLAlchemy connection string it would be the optional +driver value
    that can come after the dialect name.
    
    For example:
    
    .. code-block:: ini
    
        dialect=postgresql
        driver=psycopg2
        
    Would generate a connection string like:
    
    .. code-block:: text
    
        postgresql+psycopg2://user:password@host:port/dbname
        
    You could also provide both dialect and driver in the dialect field like this:
    
    .. code-block:: ini
    
        dialect=postgresql+psycopg2                        
    """

    host: Optional[str] = None
    """
    Hostname or IP number. May also be a data source name for some drivers.
    https://docs.sqlalchemy.org/en/20/core/engines.html#sqlalchemy.engine.URL.host
    """

    port: Optional[int] = None
    """
    Integer port number.
    https://docs.sqlalchemy.org/en/20/core/engines.html#sqlalchemy.engine.URL.port
    """

    database_name: str
    """
    https://docs.sqlalchemy.org/en/20/core/engines.html#sqlalchemy.engine.URL.database
    """

    use_get_cluster_credentials: bool = False
    """
    For Amazon Redshift only.  Should the boto3 call be performed to get temporary 
    database credentials?  If so, then rs_db_user_id is required.
    """

    rs_new_credentials_seconds: int = 1800
    """
    For Amazon Redshift and with use_get_cluster_credentials = True only.
    How long should we wait before requesting new temporary credentials last (in seconds).
    This should be less than the value used in rs_duration_seconds.
    Note: The server might respond with an even shorter duration, if so that will be used.  
    """

    rs_duration_seconds: Optional[int] = 3600
    """
    For Amazon Redshift and with use_get_cluster_credentials = True only.
    The value here is passed to the boto3 get_cluster_credentials call.   
    """

    rs_region_name: Optional[str] = None
    """
    For Amazon Redshift only. The region_name to use.
    """

    rs_cluster_id: Optional[str] = None
    """
    For Amazon Redshift only. The cluster name to use.
    """

    rs_auto_create: Optional[bool] = None
    """
    For Amazon Redshift and with use_get_cluster_credentials = True only.
    The value here is passed to the boto3 get_cluster_credentials call.   
    """

    rs_db_groups: Optional[List[str]] = None
    """
    For Amazon Redshift and with use_get_cluster_credentials = True only.
    The value here is passed to the boto3 get_cluster_credentials call.   
    """

    aws_access_key_id: Optional[str] = None
    """
    For Amazon Redshift and with use_get_cluster_credentials = True only.
    If provided, use this access key to authenticate for the use_get_cluster_credentials call.
    If not provided, the user_id value will be used instead.   
    """

    aws_secret_access_key: Optional[str] = None
    """
    For Amazon Redshift and with use_get_cluster_credentials = True only.
    If provided, use this access key to authenticate for the use_get_cluster_credentials call.
    If not provided, get_password() will be used to get the password from the configured 
    password source (see `PasswordSource` for options.)   
    """

    rs_db_user_id: Optional[str] = None
    """
    For Amazon Redshift and with use_get_cluster_credentials = True only.
    The value here is passed to the boto3 get_cluster_credentials call.   
    """

    create_engine_args: Dict[str, Any] = {}
    """
    A dictionary with extra arguments to pass to the SQLAlchemy create_engine function.
    See
    https://docs.sqlalchemy.org/en/20/core/engines.html#sqlalchemy.create_engine    
    """

    arraysize: Optional[int] = 5000
    """
    Only used for Oracle. Passed to SQLAlchemy create_engine function.
    """

    encoding: Optional[str] = None
    """
    Passed to SQLAlchemy create_engine function.
    """

    poolclass: Optional[str] = None
    """
    Pool subclass that should be passed to SQLAlchemy create_engine function.
    Only supports None, QueuePool or NullPool.
    See
    https://docs.sqlalchemy.org/en/20/core/engines.html#sqlalchemy.create_engine.params.poolclass
    """

    # Private attribute used to hold the engine
    _engine = PrivateAttr(default=None)
    _sqlite_connection = PrivateAttr(default=None)
    # Private attribute used to hold the redshift client
    _rs_client: 'RedshiftClient' = PrivateAttr(default=None)
    _rs_credential_expiry: datetime = PrivateAttr(default=None)

    # Values to hide from config exports
    _private_value_atts = PrivateAttr(default={'password', 'aws_secret_access_key'})

    def __repr__(self):
        return Credentials.__repr__(self)

    def __str__(self):
        return str(self.get_uri())

    # Note the order of decorators matters!
    # noinspection PyNestedDecorators
    @model_validator(mode="before")
    @classmethod
    def translate(cls, values):
        if not isinstance(values, dict):
            raise ValueError(f"translate values: {values} must be a dict not {type(values)}")
        
        # Convert from old setting names to new names
        name_map = {
            'dsn': 'host',
            'dbname': 'database_name',
            'key_ring_system': 'keyring_section',
            'default_user_id': 'user_id',
        }
        for old, new in name_map.items():
            if old in values and new not in values:
                warnings.warn(f"Config SQLAlchemyDatabase.{old} is deprecated. Use {new} instead.")
                values[new] = values[old]
                del values[old]

        if values.get('dialect') in {'sqlite'}:
            values['user_id'] = 'NA'
            values['host'] = 'NA'
        else:
            if values.get('host', None) is None:
                raise ValueError("host is required for databases other than sqlite")

        if 'create_engine_args' in values:
            create_engine_args = values['create_engine_args']
            if not isinstance(create_engine_args, dict):
                create_engine_args_dict = dict()
                for arg in create_engine_args.split('\n'):
                    arg = arg.strip()
                    if arg != '':
                        arg_name, arg_value = arg.split('=')
                        arg_name = arg_name.strip()
                        arg_value = arg_value.strip()
                        try:
                            arg_value_int = int(arg_value)
                            arg_value = arg_value_int
                        except ValueError:
                            pass
                        create_engine_args_dict[arg_name] = arg_value
                values['create_engine_args'] = create_engine_args_dict

        return values

    @model_validator(mode='after')
    def check_model(self):
        if self.dialect == 'sqlite':
            # Bypass password checks
            return self
        else:
            if self.use_get_cluster_credentials:
                if self.rs_cluster_id is None:
                    raise ValueError("rs_cluster_id required when use_get_cluster_credentials is True")
            # noinspection PyCallingNonCallable
            return Credentials.check_model(self)

    def get_password(self) -> str:
        try:
            return super().get_password()
        except ValueError:
            if self.dialect not in {'sqlite'}:
                raise
            else:
                # noinspection PyTypeChecker
                return None

    @staticmethod
    def _now():
        return datetime.now(timezone.utc)

    def get_cluster_credentials(self) -> Tuple[str, str]:
        if self._rs_client is None:
            import boto3

            if self.password_source == PasswordSource.AWS_ASSUME_ROLE:
                from config_wrangler.config_templates.aws.aws_session import AWS_Session
                rs_session = AWS_Session(
                    region_name=self.rs_region_name,
                    password_source=self.password_source,
                )
                self._rs_client = rs_session.session.client(
                    'redshift',
                    region_name=self.rs_region_name,
                )
            else:
                self._rs_client = boto3.client(
                    'redshift',
                    region_name=self.rs_region_name,
                    aws_access_key_id=self.aws_access_key_id,
                    aws_secret_access_key=self.aws_secret_access_key,
                    aws_session_token=None,
                )

        # Send optional params to get_cluster_credentials only if they have real values
        extra_get_cluster_credentials_args = dict()
        if self.rs_auto_create is not None:
            extra_get_cluster_credentials_args['AutoCreate'] = self.rs_auto_create

        if self.rs_db_groups is not None:
            extra_get_cluster_credentials_args['DbGroups'] = self.rs_db_groups

        new_credentials = self._rs_client.get_cluster_credentials(
                DbUser=self.rs_db_user_id,
                DbName=self.database_name,
                DurationSeconds=self.rs_duration_seconds,
                ClusterIdentifier=self.rs_cluster_id,
                **extra_get_cluster_credentials_args
            )
        rs_new_credentials_timedelta = timedelta(seconds=self.rs_new_credentials_seconds)
        self._rs_credential_expiry = self._now() + rs_new_credentials_timedelta
        if 'DbUser' not in new_credentials or 'DbPassword' not in new_credentials:
            raise SQLAlchemyError(f'Invalid response from get_cluster_credentials got {new_credentials}')
        if 'Expiration' in new_credentials:
            # Expire when redshift said or sooner as specified by rs_new_credentials_seconds
            server_expiry = new_credentials['Expiration']
            if server_expiry.tzinfo is None:
                # Moto seems to return naive datetime in local tz.
                # Boto3 returns tz aware datetime
                server_expiry = server_expiry.astimezone(tz=None)
            self._rs_credential_expiry = min(self._rs_credential_expiry, server_expiry)

        return new_credentials['DbUser'], new_credentials['DbPassword']

    def _get_connector(self) -> str:
        if self.driver is None:
            return self.dialect
        else:
            return f"{self.dialect}+{self.driver}"

    def get_uri(self) -> URL:
        user_id = self.user_id
        if self.use_get_cluster_credentials:
            if self.password_source != PasswordSource.AWS_ASSUME_ROLE:
                if self.aws_access_key_id is None:
                    self.aws_access_key_id = self.user_id
                if self.aws_secret_access_key is None:
                    self.aws_secret_access_key = self.get_password()

            if self.rs_db_user_id is None:
                raise ValueError(
                    f'rs_db_user_id required for {self.__class__.__name__} {self.host} '
                    f'with use_get_cluster_credentials = True.'
                )

            user_id, password = self.get_cluster_credentials()
        else:
            # Will be set below
            password = None

        if self.dialect in {'sqlite'}:
            return URL.create(
                drivername=self._get_connector(),
                database=self.database_name,
            )
        else:
            if not self.use_get_cluster_credentials:
                password = self.get_password()

            try:
                return URL.create(
                    drivername=self._get_connector(),
                    host=self.host,
                    port=self.port,
                    username=user_id,
                    password=password,
                    database=self.database_name,
                )
            except AttributeError:
                return URL(
                    drivername=self._get_connector(),
                    host=self.host,
                    port=self.port,
                    username=user_id,
                    password=password,
                    database=self.database_name,
                    query={},
                )

    def get_engine(self) -> Engine:
        if self._engine is None:

            kwargs = self.create_engine_args or {}
            if self.dialect == 'oracle':
                if 'arraysize' not in kwargs:
                    kwargs['arraysize'] = self.arraysize
            if self.encoding:
                kwargs['encoding'] = self.encoding
            if self.poolclass:
                if self.poolclass == 'QueuePool':
                    kwargs['poolclass'] = QueuePool
                elif self.poolclass == 'NullPool':
                    kwargs['poolclass'] = NullPool
                else:
                    raise ValueError(f'Unexpected poolclass {self.poolclass}')

            self._engine = create_engine(self.get_uri(), **kwargs)

            # Make an event listener, so we can get new RS credentials if needed
            @event.listens_for(self._engine, 'do_connect', named=True)
            def engine_do_connect(**kw):
                """
                listen for the 'do_connect' event
                """
                # from bi_etl.utility import dict_to_str
                # print(dict_to_str(kw))
                if self.use_get_cluster_credentials:
                    if self._now() >= self._rs_credential_expiry:
                        logging.info('Getting new Redshift cluster credentials')
                        db_user, password = self.get_cluster_credentials()
                        kw['cparams']['user'] = db_user
                        kw['cparams']['password'] = password

                # Return None to allow control to pass to the next event handler and ultimately
                # to allow the dialect to connect normally, given the updated arguments.
                return None
            # End engine_do_connect sub function

        return self._engine

    def raw_connection(self):
        """
        Return a "raw" DBAPI connection from the connection pool.

        See https://docs.sqlalchemy.org/en/20/core/connections.html#sqlalchemy.engine.Engine.raw_connection
        """
        return self.get_engine().raw_connection()

    def connect(self) -> 'sqlalchemy.engine.base.Connection':
        """
        Connect to the configured database.
        Special handling for sqlite where subsequent calls will return the same connection.
        """
        if self.dialect == 'sqlite':
            if self._sqlite_connection is None or self._sqlite_connection.closed:
                self._sqlite_connection = self.get_engine().connect()
            return self._sqlite_connection
        else:
            return self.get_engine().connect()

    def session(self) -> Session:
        """
        Build a SQLAlchemy session from the configured database.
        """
        return Session(bind=self.get_engine())



from __future__ import annotations

import json
import warnings
from typing import MutableMapping, Any, TYPE_CHECKING, List, Dict, Set, Generator, Tuple, Literal

from pydantic import PrivateAttr, BaseModel, ValidationError

from config_wrangler.config_wrangler_config import ConfigWranglerConfig

# noinspection PyProtectedMember

if TYPE_CHECKING:
    from config_wrangler.config_from_loaders import ConfigFromLoaders


# class SectionMissingError(PydanticValueError):
#     msg_template = 'Section required'
#
#
# class BadValueError(PydanticValueError):
#     msg_template = '{original}. value_provided = {value_str}'
#
#     def __init__(self, original: PydanticValueError, value_str: str) -> None:
#         super().__init__(original=original, value_str=value_str)
#

private_attrs = ('_root_config', '_parents', '_name_map')


class ConfigHierarchy(BaseModel):
    """
    A non-root member of a hierarchy of configuration items.

    NOTE: This class requires that the top of the hierarchy be an instance of
    :py:class:`config_wrangler.config_root.ConfigRoot`
    """
    model_config = ConfigWranglerConfig(
        validate_default=True,
        validate_assignment=True,
        validate_credentials=True
    )
    _root_config: 'ConfigFromLoaders' = PrivateAttr(default=None)
    _DEFAULT_PARENTS = ['parents_not_set']
    _parents: List[str] = PrivateAttr(default=_DEFAULT_PARENTS)
    _name_map: Dict[str, str] = PrivateAttr(default={})
    _private_value_atts = PrivateAttr(default={})

    # noinspection PyMethodParameters
    # noinspection PyProtectedMember
    def __init__(__pydantic_self__, **data: Any) -> None:
        """
        Create a new model by parsing and validating input data from keyword arguments.

        Raises ValidationError if the input data cannot be parsed to form a valid model.

        Uses something other than `self` the first arg to allow "self" as a settable attribute
        """
        private_holding = dict()
        for attr in private_attrs:
            if attr in data:
                private_holding[attr] = data[attr]
                del data[attr]
        try:
            super().__init__(**data)
        except ValidationError as e:
            # Limit the depth of the traceback
            for error in e.errors():
                try:
                    location = error['ctx']['location']
                    location_str = '.'.join(location)
                    input_data = error['ctx']['input']
                    # Input might not be a dict
                    for setting, value in e.errors()['ctx']['input']:
                        print(f"input dict for {location_str}: {setting}={value}")
                except Exception:
                    pass
            raise ValidationError.from_exception_data(
                title=e.title,
                line_errors=e.errors(),
            ) from None
        finally:
            pass
        for attr, attr_value in private_holding.items():
            setattr(__pydantic_self__, attr, attr_value)

    def _private_attr_dict(self) -> dict:
        return {
            attr: getattr(self, attr, None) for attr in private_attrs
        }

    def _dict_for_init(
            self,
            exclude: Set[str] = None,
    ) -> Dict[str, Any]:
        d = self.model_dump()
        d.update(**self._private_attr_dict())
        if exclude is not None:
            for exclude_attr in exclude:
                if exclude_attr in d:
                    del d[exclude_attr]
        return d

    def full_item_name(self, item_name: str = None, delimiter: str = ' -> '):
        """
        The fully qualified name of this config item in the config hierarchy.
        """
        if self._parents == self._DEFAULT_PARENTS:
            parents = [self.__class__.__name__]
        else:
            parents = self._parents

        if item_name is None:
            return delimiter.join(parents)
        else:
            return delimiter.join(parents + [item_name])

    @staticmethod
    def translate_config_data(config_data: MutableMapping):
        """
        Children classes can provide translation logic to allow older config files to be used
        with newer config class definitions.
        """
        return config_data

    def get(self, section, item, fallback=...):
        """
        Used as a drop in replacement for ConfigParser.get() with dynamic config field names
        (using a string variable for the section and item names instead of python code attribute access)

        .. warning::

            With this method Python code checkers (linters) will not warn about invalid config items.
            You can end up with runtime AttributeError errors.
        """
        try:
            section_obj = getattr(self, section)
            return getattr(section_obj, item)
        except AttributeError:
            if fallback is ...:
                raise
            else:
                return fallback

    def getboolean(self, section, item, fallback=...) -> bool:
        """
        Used as a drop in replacement for ConfigParser.getboolean() with dynamic config field names
        (using a string variable for the section and item names instead of python code attribute access)

        .. warning::

            With this method Python code checkers (linters) will not warn about invalid config items.
            You can end up with runtime AttributeError errors.
        """
        value = self.get(section=section, item=item, fallback=fallback)
        if value is None:
            value = False
        if not isinstance(value, bool):
            raise ValueError('getboolean called on non-bool config item')
        return value

    def get_list(self, section, item, fallback=...) -> list:
        """
        Used as a drop in replacement for ConfigParser.get() + list parsing with dynamic config field names
        (using a string variable for the section and item names instead of python code attribute access)
        that is then parsed as a list.

        .. warning::

            With this method Python code checkers (linters) will not warn about invalid config items.
            You can end up with runtime AttributeError errors.
        """
        value = self.get(section=section, item=item, fallback=fallback)
        if value is None:
            value = []
        if not isinstance(value, list):
            raise ValueError('get_list called on non-list config item')
        return value

    def __getitem__(self, section):
        try:
            section_obj = getattr(self, section)
            return section_obj.dict()
        except AttributeError as e:
            raise KeyError(str(e))

    def add_child(self, name: str, child_object: 'ConfigHierarchy'):
        """
        Set this configuration as a child in the hierarchy of another config.
        For any programmatically created config objects this is required so that the
        new object 'knows' where it lives in the hierarchy -- most importantly so that
        it can find the hierarchies root object.
        """
        child_object._parents = self._parents + [name]
        child_object._root_config = self._root_config

    def set_as_child(self, name: str, other_config_item: 'ConfigHierarchy'):
        warnings.warn(
            'The `set_as_child` method is deprecated; use `add_child` instead.',
            DeprecationWarning,
            stacklevel=2,
        )

        self.add_child(name, other_config_item)

    def get_copy(self, copied_by: str = 'get_copy') -> 'ConfigHierarchy':
        """
        Copy this configuration. Useful when you need to programmatically modify a
        configuration without modifying the original base configuration.
        """
        new_instance = self.model_copy(deep=False)
        try:
            self.add_child(copied_by, new_instance)
        except AttributeError:
            # Make the copy its own root
            new_instance._root_config = new_instance
            new_instance._parents = []
        return new_instance

    def __iter__(self) -> Generator[Tuple[str, Any], None, None]:
        """
        Iterate through the values, but hide any password (_private_value_atts)
        values in this output.  Passwords should be directly accessed by attribute name.
        """
        for key, value in super().__iter__():
            if key in self._private_value_atts and value is not None:
                value = '*' * 8
            yield key, value

    def model_dump_non_private(
            self,
            *,
            mode: Literal['json', 'python'] | str = 'python',
            exclude: Set[str] = None,
    ) -> dict[str, Any]:

        result = dict()

        if exclude is None:
            exclude = {}

        for key, value in self:
            if key not in exclude:
                if isinstance(value, ConfigHierarchy):
                    result[key] = value.model_dump_safe(mode=mode, exclude=exclude)
                else:
                    if mode == 'json':
                        result[key] = json.dumps(value)
                    else:
                        result[key] = value
        return result



from collections import defaultdict
from types import MethodType
from typing import Set

from pydantic import BaseModel

class_registry = defaultdict(set)


def get_class_that_defined_method(method: MethodType) -> str:
    parts = method.__qualname__.split('.')
    return '.'.join(parts[:-1])


def config_hierarchy_validator(validation_fn: MethodType):
    # validation_fn._is_config_hierarchy_validator = True

    def decorated_validator(*args, **kwargs):
        validation_fn(*args, **kwargs)

    decorated_validator._is_config_hierarchy_validator = True
    return decorated_validator


def get_validation_functions(model_level: BaseModel) -> Set[str]:
    return class_registry.get(model_level.__class__.__qualname__, set())



from enum import auto
from typing import Annotated, Any

from strenum import StrEnum


class PasswordSource(StrEnum):
    CONFIG_FILE = auto()
    ENVIRONMENT = auto()
    KEYRING = auto()
    KEEPASS = auto()
    AWS_ASSUME_ROLE = auto()


def _check_password_source(value: Any) -> PasswordSource:
    if isinstance(value, str):
        value = value.upper()
    return PasswordSource[value]


PasswordSourceValidated = Annotated[
    PasswordSource,
    _check_password_source,
]


from typing import *

from config_wrangler.config_from_loaders import ConfigFromLoaders
from config_wrangler.config_data_loaders.ini_config_data_loader import IniConfigDataLoader


class ConfigFromIni(ConfigFromLoaders):
    # noinspection PyMethodParameters
    def __init__(
            __pydantic_self__,
            file_name: str = 'config.ini',
            start_path: Optional[str] = None,
            **kwargs: Dict[str, Any]
    ) -> None:
        """
        Note: Uses something other than `self` the first arg to allow "self" as a settable attribute
        """
        ini_loader = IniConfigDataLoader(start_path=start_path, file_name=file_name)
        super().__init__(
            _config_data_loaders=[ini_loader],
            **kwargs
        )



import os
from copy import deepcopy
from pathlib import Path
from typing import *

from pydantic import BaseModel
from pydicti import Dicti

from config_wrangler.config_data_loaders.base_config_data_loader import BaseConfigDataLoader
from config_wrangler.utils import merge_configs, match_config_data_to_model


class FileConfigDataLoader(BaseConfigDataLoader):
    config_inheritance_section: str = 'Config'
    config_inheritance_field_name_prefix: str = 'parent'

    def __init__(
            self,
            file_name: str,
            start_path: Optional[Union[str, Path]] = None,
    ):
        super().__init__()
        self.start_path = start_path or os.getcwd()
        self.file_name = file_name
        self.files_read = []

    def _read_file(self, file_path: Path) -> MutableMapping:
        raise NotImplementedError()

    def _read_file_case_insensitive(self, file_path: Path) -> Dicti:
        return Dicti(self._read_file(file_path))

    def _merge_files_into_config_data(
            self, config_data: MutableMapping,
            path: Path,
            file_name: str,
            fail_on_does_not_exist: bool = True
    ):
        full_path = Path(file_name)
        if not full_path.is_absolute():
            full_path = Path(path, file_name)

        if full_path.exists():
            file_config_data = Dicti(self._read_file_plus_inherited(full_path))
            # Check for and remove any [Config] parent settings.
            # They should have already been used, but we don't want to merge them up
            if self.config_inheritance_section in file_config_data:
                for field_name in list(file_config_data[self.config_inheritance_section]):
                    if field_name.startswith(self.config_inheritance_field_name_prefix):
                        del file_config_data[self.config_inheritance_section][field_name]
            merge_configs(config_data, file_config_data)
        elif fail_on_does_not_exist:
            raise FileNotFoundError(f"{file_name} not found with path = {path}")
        return config_data

    def _check_inherited_files(
            self,
            config_data: Dicti,
            path: Path,

    ) -> Dicti:
        if self.config_inheritance_section in config_data:
            for field_name in config_data[self.config_inheritance_section]:
                if field_name.startswith(self.config_inheritance_field_name_prefix):
                    file_name = config_data[self.config_inheritance_section][field_name]
                    if '\n' in file_name:
                        for file_name_part in file_name.split('\n'):
                            self._merge_files_into_config_data(config_data, path, file_name_part)
                    else:
                        self._merge_files_into_config_data(config_data, path, file_name)
        return config_data

    def _read_file_plus_inherited(self, file_path: Path):
        file_config_data = self._read_file_case_insensitive(file_path)
        folder = file_path.parents[0]
        with_parents_added = self._check_inherited_files(file_config_data, path=folder)
        return with_parents_added

    def read_config_data(self, model: BaseModel) -> MutableMapping:
        full_path = Path(self.start_path, self.file_name)

        config_data = deepcopy(self._init_config_data)

        if full_path.exists():
            file_config_data = self._read_file_plus_inherited(full_path)
            merge_configs(config_data, file_config_data)

        for parent_dir in full_path.parents:
            parent_path = Path(parent_dir, self.file_name)
            if parent_path != full_path:
                if parent_path.exists():
                    file_config_data = self._read_file_plus_inherited(parent_path)
                    merge_configs(config_data, file_config_data)

        if len(self.files_read) == 0:
            raise FileNotFoundError(f"{full_path} or {self.file_name} in parent directories")

        match_config_data_to_model(model, config_data)
        return config_data



import logging
import os
from copy import deepcopy
from typing import *

from pydantic import BaseModel

from config_wrangler.config_data_loaders.base_config_data_loader import BaseConfigDataLoader
from config_wrangler.utils import walk_model, match_config_data_to_field, has_sub_fields


class EnvConfigDataLoader(BaseConfigDataLoader):
    """
    Copied initial version from Pydantic BaseSettings.
    Made to fit into ConfigLoader approach with nested models.
    """

    def __init__(self, env_prefix: str = None, config_data_dict: Dict[str, Any] = None, **kwargs):
        super().__init__(config_data_dict=config_data_dict, **kwargs)
        self.env_prefix = env_prefix or ''
        self.log = logging.getLogger(__name__)

    def read_config_data(self, model: BaseModel) -> MutableMapping:
        config_data = deepcopy(self._init_config_data)

        env_vars = {k.lower(): v for k, v in os.environ.items()}

        for field_name, field_info, parents in walk_model(model):
            env_val: Optional[str] = None
            # Note: field.field_info.extra['env_names'] is set from the 'env' variable on the field or in Config.
            #       See https://pydantic-docs.helpmanual.io/usage/settings/#environment-variable-names

            field_possible_names = [field_info.serialization_alias, field_info.alias, field_name]

            # Search for env var named directly after the prefix + field (it would apply to that field in ANY section)
            if self.env_prefix is not None and self.env_prefix != '':
                for env_name in field_possible_names:
                    if env_name is not None:
                        env_full_name = f"{self.env_prefix}{env_name}"
                        env_val = env_vars.get(env_full_name.lower())
                        if env_val is not None:
                            break

            if env_val is None:
                possible_name_delim = ('_', '.', ':')
                possible_env_vars = set()

                for field_name_option in field_possible_names:
                    for delimiter in possible_name_delim:
                        if field_name_option:
                            full_name = delimiter.join(parents + [field_name_option])
                        else:
                            full_name = delimiter.join(parents)
                        env_name = f"{self.env_prefix}{full_name}"
                        if env_name in env_vars:
                            possible_env_vars.add(env_name)
                if len(possible_env_vars) > 1:
                    raise ValueError(f"Found multiple matches for {parents} {field_name} {field_info}.  They are: {possible_env_vars}")
                elif len(possible_env_vars) == 1:
                    env_var = list(possible_env_vars)[0]
                    self.log.info(f"Read ENV {env_var} into {parents} {field_name} {field_info}")
                    env_val = env_vars[env_var]

            if has_sub_fields(field_info.annotation):
                env_val = match_config_data_to_field(
                    field_name=field_name,
                    field_info=field_info,
                    field_value=env_val,
                    parent_container={},
                    root_config_data={},
                    parents=parents,
                )

            # Walk down the hierarchy making nodes as needed
            if env_val is not None:
                sub_config = config_data
                for parent in parents:
                    if parent in sub_config:
                        sub_config = sub_config[parent]
                    else:
                        new_sub_config = dict()
                        sub_config[parent] = new_sub_config
                        sub_config = new_sub_config
                # Only set the value if it is not already in our config_data
                if field_name not in sub_config:
                    sub_config[field_name] = env_val
        return config_data



from strenum import StrEnum
from enum import auto


def auto_str():
    # noinspection PyArgumentList
    return auto()


__all__ = ['StrEnum', 'auto_str']

# Convenience module so that enum configs don't have to do imports from two locations



import os
import warnings
from typing import Optional

from pydantic import PrivateAttr, model_validator, Field

from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy
from config_wrangler.config_templates.keepass_config import KeepassConfig
from config_wrangler.config_templates.password_defaults import PasswordDefaults
from config_wrangler.config_templates.password_source import PasswordSource, PasswordSourceValidated

__all__ = ['Credentials', 'PasswordDefaults', 'PasswordSource']

from config_wrangler.validate_config_hierarchy import config_hierarchy_validator


class Credentials(ConfigHierarchy):
    user_id: Optional[str] = None
    """
    The user ID to use
    """

    password_source: Optional[PasswordSourceValidated] = None
    """
    The source to use when getting a password for the user.  
    See :py:class:`PasswordSource` for valid values.
    """

    raw_password: Optional[str] = Field(repr=False, default=None)
    """
    This is only used for the extremely non-secure `CONFIG_FILE` password source.
    The password is stored directly in the config file next to the user_id with
    the setting name `raw_password`
    """

    keyring_section: Optional[str] = None
    """
    If the password_source is KEYRING, then which section (AKA system)
    should this module look for the password in.
    
    See https://pypi.org/project/keyring/
    or https://github.com/jaraco/keyring
    """
    keepass_config: Optional[str] = 'keepass'
    """
    If the password_source is KEEPASS, then which root level config item contains
    the settings for Keepass (must be an instance of 
    :py:class:`config_wrangler.config_templates.keepass_config.KeepassConfig`)
    """
    keepass: Optional[KeepassConfig] = None
    """
    If the password_source is KEEPASS, then load a sub-section with the 
    :py:class:`config_wrangler.config_templates.keepass_config.KeepassConfig`) settings
    """

    keepass_group: Optional[str] = None
    """
    If the password_source is KEEPASS, which group in the Keepass database should
    be searched for an entry with a matching entry.
    
    If is None, then the `KeepassConfig.default_group` value will be checked.
    If that is also None, then a ValueError will be raised.
    """

    keepass_title: Optional[str] = None
    """
    If the password_source is KEEPASS, this is an optional filter on the title
    of the keepass entries in the group.
    """

    validate_password_on_load: bool = True
    """
    Should config_wrangler query the password source for this password at 
    load (startup) time? If so, it will raise an error if the password is 
    None or an empty string. It **does not** actually connect or 
    authenticate the user_id & password combination.
    """

    # Values to hide from config exports
    _private_value_atts = PrivateAttr(default={'raw_password'})

    def _get_password_keyring(self):
        if self.keyring_section is None:
            raise ValueError(f"{self.full_item_name()} password_source is KEYRING but keyring_section is None")
        import keyring
        password = keyring.get_password(self.keyring_section, self.user_id)
        search_info = f"{self.keyring_section} {self.user_id}"
        return password, search_info

    def _get_password_config(self):
        warnings.warn(
            'Passwords stored directly in config or worse in code are not safe. Please make sure to fix this before deploying.',
            stacklevel=3,
        )
        password = self.raw_password
        search_info = f"{self.full_item_name()}.raw_password"
        return password, search_info

    @staticmethod
    def _get_envt_name(search_name: str) -> Optional[str]:
        if search_name in os.environ:
            return search_name
        else:
            search_name_upper = search_name.upper()
            for envt_var in os.environ:
                envt_var_upper = envt_var.upper()
                if envt_var_upper == search_name_upper:
                    return envt_var
            return None

    def _get_password_environment(self):
        password = self.raw_password
        if password is not None and password != '':
            # Env loader might have already found it for us. At least we'll assume that's where it came from.
            search_info = f"{self.full_item_name()}.raw_password"
            return password, search_info
        else:
            name_in_envt = self.full_item_name(delimiter='_').replace('.', '_')
            search_info_list = []
            password = None
            for envt_name_search in [
                f"PASSWORD_{self.user_id}",
                f"{self.user_id}_PASSWORD",
                self.user_id,
                f"{name_in_envt}_PASSWORD",
            ]:
                search_info_list.append(envt_name_search)
                envt_name_found = self._get_envt_name(envt_name_search)
                if envt_name_found is not None:
                    if envt_name_found in os.environ:
                        password = os.environ[envt_name_found]
                        break
            if password is None:
                raise ValueError(f"{self.full_item_name()} password_source = ENVIRONMENT. Value not found in {search_info_list}")
            else:
                return password, ','.join(search_info_list)

    def _get_keepass_config_str_ref(self) -> KeepassConfig:
        # Try older keepass_config string reference
        keepass_config = None
        if self.keepass_config is not None:
            try:
                if self._root_config is None:
                    raise ValueError("get_password called on Credentials that are not part of a ConfigRoot hierarchy (and keepass_config used)")
                keepass_config = getattr(self._root_config, self.keepass_config)
                if not isinstance(keepass_config, KeepassConfig):
                    raise ValueError(f"{self.full_item_name()} keepass_config is {type(keepass_config)} = {keepass_config} not KeepassConfig instance")
            except (KeyError, AttributeError):
                pass
        return keepass_config

    def _get_keepass_config_sub_section(self) -> KeepassConfig:
        if self.keepass is not None:
            keepass_config = self.keepass
        else:
            # Try root passwords subsection definition
            if self._root_config is None:
                raise ValueError(
                    "get_password called on Credentials that are not part of a ConfigRoot hierarchy "
                    "and does not have a local keepass_config or keepass sub-section"
                )
            if self._root_config.passwords is None:
                raise ValueError(
                    f"{self.full_item_name()} get_password called with PasswordSource = KEEPASS, "
                    f"but keepass_config/keepass is not in "
                    f"local section and yet the root passwords section is missing."
                )
            if self._root_config.passwords.keepass is None:
                raise ValueError(
                    f"{self.full_item_name()} get_password called with PasswordSource = KEEPASS, but keepass_config/keepass is not in "
                    f"local section and yet the root passwords section is also missing keepass_config."
                )
            keepass_config = self._root_config.passwords.keepass

        if not isinstance(keepass_config, KeepassConfig):
            raise ValueError(f"{self.full_item_name()} keepass_config is {type(keepass_config)} = {keepass_config} not KeepassConfig instance")

        return keepass_config

    def _get_password_keepass(self):
        keepass_config = self._get_keepass_config_str_ref()
        if keepass_config is None:
            keepass_config = self._get_keepass_config_sub_section()

        try:
            search_info = keepass_config.full_item_name()
            get_password = keepass_config.get_password
            try:
                password = get_password(self.keepass_group, self.keepass_title, self.user_id)
            except ValueError as e:
                raise ValueError(f"{self.full_item_name()} -> keepass config {keepass_config.full_item_name()} error {e}")
        except AttributeError as e:
            raise ValueError(
                f"{self.full_item_name()} -> keepass config {keepass_config.full_item_name()} does not appear to be valid. "
                f"{e}"
            )

        return password, search_info

    def get_password(self) -> str:
        """
        Get the password for this resource.
        `password_source` controls where it looks for the password.
        If that is None, then the root level `passwords` container is checked for `password_source` value.
        """
        if self.password_source is None:
            if self._root_config is None:
                raise ValueError("get_password called on Credentials that are not part of a ConfigRoot hierarchy")
            try:
                passwords_defaults = getattr(self._root_config, 'passwords')
            except AttributeError:
                raise ValueError(
                    f"{self.full_item_name()} password_source not provided and 'passwords' section not found"
                )
            if passwords_defaults is None:
                raise ValueError(
                    f"{self.full_item_name()} password_source not provided "
                    f"and 'passwords' section does not exist."
                )
            else:
                try:
                    self.password_source = passwords_defaults.password_source
                except AttributeError as e:
                    raise ValueError(
                        f"{self.full_item_name()} password_source not provided "
                        f"and 'passwords' section does not have 'password_source' {e} "
                    )

        try:
            if self.password_source == PasswordSource.KEYRING:
                password, search_info = self._get_password_keyring()
            elif self.password_source == PasswordSource.CONFIG_FILE:
                password, search_info = self._get_password_config()
            elif self.password_source == PasswordSource.ENVIRONMENT:
                password, search_info = self._get_password_environment()
            elif self.password_source == PasswordSource.KEEPASS:
                password, search_info = self._get_password_keepass()
            elif self.password_source == PasswordSource.AWS_ASSUME_ROLE:
                password = 'NOT A REAL PASSWORD'
                password_source = 'Assume Role'
            else:
                raise ValueError(f"invalid password_source")
        except Exception as e:
            raise ValueError(f"{self.full_item_name()} with password_source={self.password_source} got error {e}")

        if password is None or password == '':
            raise ValueError(f"{self.full_item_name()} password is not set. Source is {self.password_source} location = {search_info}")

        return password

    @model_validator(mode='after')
    def check_model(self):
        user_id = self.user_id
        if user_id == '' or user_id is None:
            if self.password_source != PasswordSource.AWS_ASSUME_ROLE:
                raise ValueError("user_id not provided")

        if self.password_source == PasswordSource.KEYRING:
            keyring_section = self.keyring_section
            if keyring_section is None:
                raise ValueError(f"{self.full_item_name()} password_source is KEYRING but keyring_section is None")
        elif self.password_source == PasswordSource.CONFIG_FILE:
            password = self.raw_password
            if password is None or password == '':
                raise ValueError(f"{self} password_source is Config but password is not set")
        # elif self.password_source == PasswordSource.KEEPASS:
        # TODO: Need to check _get_keepass_config_sub_section() and local model for keepass_group
        return self

    @config_hierarchy_validator
    def _check_config_hierarchy(self):
        """
        Validator for checking that credentials are valid.
        The validation is performed if all of these are True:
        - validate_credentials is True or not set in the root model_config object.
        - validate_default is True or not set in the root model_config object.
        - validate_password_on_load is True (defaults to True) in this Credentials based object.
        """
        if self._root_config is None:
            raise ValueError("Credentials are not part of a ConfigRoot hierarchy")
        else:
            model_config = self._root_config.model_config
            validate_credentials = getattr(model_config, 'validate_credentials', True)
            if model_config.get('validate_default', True) and validate_credentials and self.validate_password_on_load:
                root_config = self._root_config.model_config
                root_validate_credentials = root_config.get('validate_credentials', True)
                if root_validate_credentials:
                    _ = self.get_password()



import os
import shutil
from pathlib import Path
from typing import Any

from pydantic import DirectoryPath, FilePath, BeforeValidator
from typing_extensions import Annotated

# Make sure these can be imported from here and not just from pydantic
DirectoryPath = DirectoryPath
FilePath = FilePath


def _path_validator(value: Any) -> Path | None:
    if value is None:
        return None
    return Path(value)


def _file_validator(p: Path) -> Path:
    if p is None:
        raise ValueError(f"{p} is not a file")
    if not p.is_file():
        raise ValueError(f"{p} is not a file")
    return p


def _writable_file_validator(p: Path) -> Path:
    if p is None:
        raise ValueError(f"{p} is not a file")

    if p.is_file():
        if not os.access(p, os.W_OK):
            raise ValueError(f"{p} exists and is not writable")
    else:
        parent = p.parent
        if not parent.is_dir():
            raise ValueError(f"{p} error {parent} is not a directory")
        else:
            if not os.access(parent, os.W_OK):
                raise ValueError(f"{p.name} can't be created in {parent}")
    return p


def _directory_validator(p: Path) -> Path:
    if p is None:
        raise ValueError(f"{p} is not a directory")

    if not p.is_dir():
        raise ValueError(f"{p} is not a directory")
    return p


def _expand_user_validator(p: Path | None) -> Path | None:
    if p is None:
        return None
    return p.expanduser()


def _path_exists(p: Path) -> Path:
    if p is None:
        raise ValueError(f"Could not find path {p}")

    if not p.exists():
        raise ValueError(f"Could not find path {p}")
    return p


def _ensure_exists_validator(p: Path) -> Path:
    if p is None:
        raise ValueError(f"Can't make None path")

    if not p.exists():
        try:
            p.mkdir(parents=True, exist_ok=True)
            if not p.exists():
                raise ValueError(f"Could not create directory {p}")
        except Exception as e:
            raise ValueError(f"make dir {p} yields error: {e}")
    return p


def _find_up(path: Path) -> Path:
    if path is None:
        raise ValueError(f"Can't find None path")
    if path.exists():
        return path
    else:
        start_dir = Path(os.getcwd())
        for parent_dir in start_dir.parents:
            parent_path = parent_dir / path
            if parent_path.exists():
                return parent_path
        raise ValueError(f"{path} not found in {start_dir} or parents")


def _find_in_system_path(path: Path) -> Path:
    if path is None:
        raise ValueError(f"Can't find None path")

    full_path = shutil.which(str(path))
    if full_path is None:
        raise ValueError(f"{path} not found")
    # Note: on Windows any existing file appears as executable
    elif not os.access(full_path, os.X_OK):
        raise ValueError(f"{path} found at {full_path} but is not executable")
    return Path(full_path)


WritableFile = Annotated[
    Path,
    BeforeValidator(_writable_file_validator),
    BeforeValidator(_expand_user_validator),
    BeforeValidator(_path_validator)
]

PathExpandUser = Annotated[
    Path,
    BeforeValidator(_path_exists),
    BeforeValidator(_expand_user_validator),
    BeforeValidator(_path_validator)
]


DirectoryExpandUser = Annotated[
    Path,
    BeforeValidator(_directory_validator),
    BeforeValidator(_expand_user_validator),
    BeforeValidator(_path_validator)
]


AutoCreateDirectoryPath = Annotated[
    Path,
    BeforeValidator(_directory_validator),
    BeforeValidator(_ensure_exists_validator),
    BeforeValidator(_expand_user_validator),
    BeforeValidator(_path_validator)
]


PathFindUp = Annotated[
    Path,
    BeforeValidator(_find_up),
    BeforeValidator(_path_validator)
]


DirectoryFindUp = Annotated[
    Path,
    BeforeValidator(_directory_validator),
    BeforeValidator(_find_up),
    BeforeValidator(_path_validator)
]

PathFindUpExpandUser = Annotated[
    Path,
    BeforeValidator(_find_up),
    BeforeValidator(_expand_user_validator),
    BeforeValidator(_path_validator)
]

DirectoryFindUpExpandUser = Annotated[
    Path,
    BeforeValidator(_directory_validator),
    BeforeValidator(_find_up),
    BeforeValidator(_expand_user_validator),
    BeforeValidator(_path_validator)
]


ExecutablePath = Annotated[
    Path,
    BeforeValidator(_file_validator),
    BeforeValidator(_find_in_system_path),
    BeforeValidator(_path_validator)
]



from typing import TYPE_CHECKING

from pydantic import PrivateAttr

from config_wrangler.config_templates.aws.aws_session import AWS_Session

if TYPE_CHECKING:
    # https://youtype.github.io/boto3_stubs_docs/mypy_boto3_ssm/
    from mypy_boto3_secretsmanager.client import SecretsManagerClient
    from mypy_boto3_secretsmanager.type_defs import GetSecretValueResponseTypeDef


class SecretsManager(AWS_Session):
    _service: str = PrivateAttr(default='secretsmanager')
    # Region is not optional for SecretsManager
    region_name: str

    @property
    def client(self) -> 'SecretsManagerClient':
        return super().client

    def get_secret_value_response(self, secret_id: str) -> 'GetSecretValueResponseTypeDef':
        return self.client.get_secret_value(SecretId=secret_id)

    def get_secret_value_by_id(self, secret_id: str) -> str:
        response = self.get_secret_value_response(secret_id=secret_id)
        try:
            return response['SecretString']
        except KeyError:
            raise ValueError(f"{secret_id} did not have SecretString. Maybe it's binary?")

    def get_secret_value_bytes_by_id(self, secret_id: str) -> bytes:
        response = self.get_secret_value_response(secret_id=secret_id)
        try:
            return response['SecretBinary']
        except KeyError:
            raise ValueError(f"{secret_id} did not have SecretBinary. Maybe it's string?")


class SecretValue(SecretsManager):
    secret_id: str

    def get_secret_value(self) -> str:
        return self.get_secret_value_by_id(secret_id=self.secret_id)

    def get_secret_value_bytes(self) -> bytes:
        return self.get_secret_value_bytes_by_id(secret_id=self.secret_id)



import logging
from typing import *

from config_wrangler.config_data_loaders.base_config_data_loader import BaseConfigDataLoader
from config_wrangler.config_root import ConfigRoot
from config_wrangler.utils import merge_configs, interpolate_values


class ConfigFromLoaders(ConfigRoot):
    """
    Base class for settings, allowing values to be set by files or environment variables.
    """

    # noinspection PyMethodParameters
    def __init__(
        __pydantic_self__,
        _config_data_loaders: List[BaseConfigDataLoader],
        config_load_log_level: int = logging.INFO,
        **kwargs: Dict[str, Any]
    ) -> None:
        """
        Note: Uses something other than `self` the first arg to allow "self" as a settable attribute
        """
        logging.basicConfig(level=config_load_log_level)
        log = logging.getLogger(__name__)

        config_data = dict(**kwargs)
        for loader in _config_data_loaders:
            log.debug(f"Loading config with {loader}")
            loader_config_data = loader.read_config_data(__pydantic_self__)
            merge_configs(config_data, loader_config_data)
        log.debug("Interpolating config macro references")
        interpolate_errors = interpolate_values(config_data, config_data)
        if len(interpolate_errors) > 0:
            log = logging.getLogger(__name__)
            log.error(f"{len(interpolate_errors)} Variable interpolation config errors found:")
            errors_str_list = []
            indent = ' ' * 3
            for error in interpolate_errors:
                error_msg = f"{indent}{error[0]}: {error[1]}"
                log.error(error_msg)
                errors_str_list.append(error_msg)

            errors_str = f"\n".join(errors_str_list)
            raise ValueError(f"Config Interpolation Errors (cnt={len(interpolate_errors)}). Errors=\n{indent}{errors_str}")

        log.debug("Translating config with translate_config_data method")
        config_data = __pydantic_self__.translate_config_data(config_data)
        super().__init__(**config_data)



from typing import *

from pydantic import PrivateAttr

try:
    import boto3
except ImportError:
    raise ImportError("AWS_Session requires boto3 to be installed")

if TYPE_CHECKING:
    from config_wrangler.config_templates.aws.s3_bucket import S3_Bucket, S3_Bucket_Key
    from mypy_boto3_sts import STSClient
    from mypy_boto3_sts.type_defs import PolicyDescriptorTypeTypeDef, TagTypeDef, ProvidedContextTypeDef
else:
    # Provide string values so that type hint code compiles
    STSClient = 'STSClient'
    PolicyDescriptorTypeTypeDef = 'PolicyDescriptorTypeTypeDef'
    TagTypeDef = 'TagTypeDef'
    ProvidedContextTypeDef = 'TagTypeDef'

# NOTE: If you are not seeing boto3-stubs code completion in Intellij-based IDEs,
#       please increase the intellisense filesize limit
#       e.g `idea.max.intellisense.filesize=20000` in IDE custom properties
#       (Help > Edit Custom Properties), then restart.

from config_wrangler.config_templates.credentials import Credentials
from config_wrangler.config_templates.credentials import PasswordSource


# AWSSession does not look right, so we added underscores
# noinspection PyPep8Naming
class AWS_Session(Credentials):
    # sso_session: Optional[DynamicallyReferenced] = None
    region_name: Optional[str] = None

    iam_role: Optional[str] = None

    _session: boto3.session.Session = PrivateAttr(default=None)
    _service: str = PrivateAttr(default=None)

    @property
    def session(self) -> boto3.session.Session:
        if self._session is None:
            if self.password_source == PasswordSource.AWS_ASSUME_ROLE:
                # Inherit role / credentials from the current environment
                args = dict()
                if self.region_name is not None:
                    args['region_name'] = self.region_name
                self._session = boto3.session.Session(
                    **args
                )
                if self.iam_role is not None:
                    self.assume_role()
            else:
                self._session = boto3.session.Session(
                    aws_access_key_id=self.user_id,
                    aws_secret_access_key=self.get_password(),
                    region_name=self.region_name,
                )
        return self._session

    def set_session(self, session: boto3.session.Session):
        self._session = session

    @property
    def has_session(self) -> bool:
        return self._session is not None

    def _get_client(self, service: str = None):
        if service is None:
            service = self._service
        # noinspection PyTypeChecker
        return self.session.client(service, region_name=self.region_name)

    @property
    def client(self):
        return self._get_client()

    def get_service_client(self, service: str):
        return self._get_client(service=service)

    def _sts_assume_role(
            self,
            role_arn: str,
            role_session_name: str,
            policy_arns: Sequence[PolicyDescriptorTypeTypeDef] = ...,
            policy: str = ...,
            duration_seconds: int = ...,
            tags: Sequence[TagTypeDef] = ...,
            transitive_tag_keys: Sequence[str] = ...,
            external_id: str = ...,
            serial_number: str = ...,
            token_code: str = ...,
            source_identity: str = ...,
            provided_contexts: Sequence[ProvidedContextTypeDef] = ...
    ) -> boto3.session.Session:
        sts_client = self.get_service_client('sts')  # type: STSClient
        response = sts_client.assume_role(
            RoleArn=role_arn,
            RoleSessionName=role_session_name,
            PolicyArns=policy_arns,
            Policy=policy,
            DurationSeconds=duration_seconds,
            Tags=tags,
            TransitiveTagKeys=transitive_tag_keys,
            ExternalId=external_id,
            SerialNumber=serial_number,
            TokenCode=token_code,
            SourceIdentity=source_identity,
            ProvidedContexts=provided_contexts,
        )

        session = boto3.session.Session(
            aws_access_key_id=response['Credentials']['AccessKeyId'],
            aws_secret_access_key=response['Credentials']['SecretAccessKey'],
            aws_session_token=response['Credentials']['SessionToken']
        )

        return session

    def assume_role(self):
        self._session = self._sts_assume_role(
            role_arn=self.iam_role,
        )

    def _get_resource(self, service: str = None):
        if service is None:
            service = self._service
        # noinspection PyTypeChecker
        return self.session.resource(service, region_name=self.region_name)

    @property
    def resource(self):
        return self._get_resource()

    def get_service_resource(self, service: str):
        return self._get_resource(service=service)

    def get_copy(self, copied_by: str = 'get_copy') -> 'AWS_Session':
        return self._factory(cls=self.__class__)

    def _factory(self, cls, exclude: Set[str] = None, **attributes):
        if exclude is None:
            exclude = set()
        exclude.update(attributes.keys())
        new_object = cls.model_construct(
            **attributes,
            **self._dict_for_init(exclude=exclude)
        )
        self.add_child(str(cls), new_object)
        if self.has_session:
            new_object.set_session(self.session)
        return new_object

    def nav_to_bucket(self, bucket_name) -> 'S3_Bucket':
        from config_wrangler.config_templates.aws.s3_bucket import S3_Bucket
        return self._factory(
            S3_Bucket,
            exclude={'bucket_name', 'folder', 'file_name', 'key'},
            bucket_name=bucket_name,
        )

    @staticmethod
    def split_s3_uri(s3_uri: str) -> Tuple[str, str]:
        parts = s3_uri.split("/", 3)
        # Note 'S3://bucket-name/key-name/file.txt'.split('/', 3)
        # Returns [ 'S3:', '', 'bucket-name', 'key-name/file.txt' ]
        if parts[0].lower() != 's3:' or parts[1] != '':
            raise ValueError(f"S3 URI '{s3_uri}' does not appear to be valid")
        if len(parts) == 3:
            return parts[2], ''
        else:
            return parts[2], parts[3]

    def nav_to_s3_link(self, s3_uri: str) -> 'S3_Bucket_Key':
        bucket, key = self.split_s3_uri(s3_uri)
        return self.nav_to_bucket(bucket) / key

    def get_ssm(self):
        from config_wrangler.config_templates.aws.ssm import SSM
        return self._factory(
            SSM,
        )

    def get_secrets_manager(self):
        from config_wrangler.config_templates.aws.secrets_manager import SecretsManager
        return self._factory(
            SecretsManager,
        )

    def get_request_v4_authorizer(self, service: str = None) -> 'requests.auth.AuthBase':
        from requests_aws4auth import AWS4Auth
        if service is None:
            service = self._service
        return AWS4Auth(
            region=self.region_name,
            service=service,
            refreshable_credentials=self.session.get_credentials(),
        )



import logging
import sys
from typing import *
from contextlib import contextmanager
from datetime import datetime, time
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from pathlib import Path

from pydantic import ByteSize, Field, model_validator
from pydicti import Dicti

from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy
from config_wrangler.config_types.enum import StrEnum, auto_str
from config_wrangler.config_types.path_types import AutoCreateDirectoryPath
from config_wrangler.utils import TZFormatter


class LogLevel(StrEnum):
    CRITICAL = auto_str()
    FATAL = auto_str()
    ERROR = auto_str()
    WARNING = auto_str()
    INFO = auto_str()
    DEBUG = auto_str()
    NOTSET = auto_str()


class FileHandlerClass(StrEnum):
    # https://docs.python.org/3/library/logging.handlers.html#rotatingfilehandler
    RotatingFileHandler = auto_str()

    # https://docs.python.org/3/library/logging.handlers.html#timedrotatingfilehandler
    TimedRotatingFileHandler = auto_str()


class LoggingConfig(ConfigHierarchy):
    console_log_level: LogLevel = LogLevel.INFO
    console_entry_format: str = '%(asctime)s - %(levelname)-8s - %(name)s: %(message)s'
    log_folder: Optional[AutoCreateDirectoryPath] = None
    log_file_name: Optional[str] = None
    add_date_to_log_file_name: Optional[bool] = None
    log_file_name_date_time_format: str = '_%Y_%m_%d_at_%H_%M_%S'
    file_log_level: LogLevel = LogLevel.DEBUG
    log_file_entry_format: str = '%(asctime)s - %(levelname)-8s - %(name)s: %(message)s'
    log_file_rotation_class: FileHandlerClass = FileHandlerClass.RotatingFileHandler

    # RotatingFileHandler specific settings
    # https://docs.python.org/3/library/logging.handlers.html#rotatingfilehandler
    log_file_max_size: ByteSize = Field(default='10 MB', validate_default=True)

    # TimedRotatingFileHandler specific settings
    # https://docs.python.org/3/library/logging.handlers.html#rotatingfilehandler
    log_file_timed_rotation_when: str = 'midnight'
    log_file_timed_rotation_interval: int = 1
    log_file_timed_rotation_attime: Optional[time] = None
    """
    Note: If log_file_timed_rotation_attime is included it must be a valid time
          with or without seconds:
             10:00
             10:30
             10:45:59
    Note 2: The time validator will accept a value 0-59 but it is interpreted as seconds 
            past midnight, not as hours like you might expect.
    """

    log_file_timed_rotation_utc: bool = False

    log_files_to_keep: int = 10
    logging_date_format: str = '%Y-%m-%d %H:%M:%S%z'
    trace_logging_setup: bool = False
    log_levels: Dict[str, LogLevel] = {}

    @model_validator(mode='after')
    def _validate_logging(self):
        if self.log_file_name is not None:
            if self.log_folder is None:
                raise ValueError(f"{self.full_item_name()} log_file_name set but no log_folder provided")
        return self

    @staticmethod
    def get_dated_log_file_name(
            log_file_prefix: str,
            date_time_format: str,
            log_file_suffix: str = '.log',
    ):
        """
        Generates a log file name with a given prefix, suffix and date time format.

        Parameters
        ----------
        log_file_prefix: str
            The part of the log file name before the date
        log_file_suffix: str
            The part of the log file name after the date
        date_time_format: str
            Optional. The date time format to use. Defaults to '_%Y_%m_%d_at_%H_%M_%S'
        """
        return f"{log_file_prefix}{datetime.now().strftime(date_time_format)}{log_file_suffix}"

    def add_log_file_handler(
            self,
            log_file_prefix: str = None,
            add_date_to_log_file_name: bool = None,
            log_file_suffix: str = '.log',
    ) -> logging.Handler:
        log = logging.getLogger(__name__)
        root_logger = logging.getLogger()
        if add_date_to_log_file_name is None:
            add_date_to_log_file_name = self.add_date_to_log_file_name
        if log_file_prefix is None:
            if self.log_file_name is not None:
                log_file_name = self.log_file_name
            else:
                log_file_name = None
        else:
            log.info(f'Using code provided prefix {log_file_prefix}')
            if add_date_to_log_file_name is None:
                if self.log_file_rotation_class == FileHandlerClass.TimedRotatingFileHandler:
                    add_date_to_log_file_name = False
                else:
                    add_date_to_log_file_name = True

            if add_date_to_log_file_name:
                log_file_name = LoggingConfig.get_dated_log_file_name(
                    log_file_prefix=log_file_prefix,
                    date_time_format=self.log_file_name_date_time_format,
                    log_file_suffix=log_file_suffix,
                )
            else:
                log_file_name = f"{log_file_prefix}{log_file_suffix}"

        file_handler = None
        if log_file_name is not None:
            if self.log_folder is not None and not Path(log_file_name).is_absolute():
                log_file_path = Path(self.log_folder, log_file_name)
            else:
                log_file_path = Path(log_file_name)

            if self.trace_logging_setup:
                log.info(f'Logging path = {log_file_path}')

            # Setup file logging

            # Make sure the directory exists
            dir_name = log_file_path.parent
            if dir_name != '.':
                dir_name.mkdir(parents=True, exist_ok=True)

            if self.log_file_rotation_class == FileHandlerClass.RotatingFileHandler:
                file_handler = RotatingFileHandler(
                    filename=log_file_path,
                    maxBytes=self.log_file_max_size,
                    backupCount=self.log_files_to_keep,
                    encoding='utf8',
                )
            elif self.log_file_rotation_class == FileHandlerClass.TimedRotatingFileHandler:
                file_handler = TimedRotatingFileHandler(
                    filename=log_file_path,
                    when=self.log_file_timed_rotation_when,
                    interval=self.log_file_timed_rotation_interval,
                    atTime=self.log_file_timed_rotation_attime,
                    utc=self.log_file_timed_rotation_utc,
                    backupCount=self.log_files_to_keep,
                    encoding='utf8',
                )
                file_handler.namer = lambda name: name.replace(".log", "") + ".log"
            else:
                raise ValueError(f"Bad log_file_rotation_class of {self.log_file_rotation_class}")

            log_file_entry_formatter = TZFormatter(self.log_file_entry_format, self.logging_date_format)
            file_handler.setFormatter(log_file_entry_formatter)
            file_handler.setLevel(self.file_log_level)
            root_logger.addHandler(file_handler)
            log.info(f"File log level = {self.file_log_level}")
        else:
            if self.trace_logging_setup:
                log.info('No log filename defined. File logging skipped.')
        return file_handler

    def remove_log_handler(self, handler: logging.Handler):
        log = logging.getLogger(__name__)
        if handler is not None:
            if self.trace_logging_setup:
                log.info(f"Closing log handler {handler}")
            root_logger = logging.getLogger()
            root_logger.removeHandler(handler)
            handler.close()
        else:
            if self.trace_logging_setup:
                log.info('No handler provided to remove_log_handler')

    @contextmanager
    def log_file_manager(
            self,
            log_file_prefix: str = None,
            add_date_to_log_file_name: bool = False,
            log_file_suffix: str = '.log',
    ):
        log_file_handler = self.add_log_file_handler(
            log_file_prefix=log_file_prefix,
            add_date_to_log_file_name=add_date_to_log_file_name,
            log_file_suffix=log_file_suffix,
        )
        try:
            yield log_file_handler
        finally:
            # Code to release resource, e.g.:
            self.remove_log_handler(log_file_handler)

    def setup_log_levels(self):
        configured_loggers = dict()
        root_logger = logging.getLogger()

        # Monkey-patch getLogger's dict to be case-insensitive        
        logger_dict = Dicti(logging.Logger.manager.loggerDict)  # @UndefinedVariable
        logging.Logger.manager.loggerDict = logger_dict

        for logger_class, desired_level_name in self.log_levels.items():
            if logger_class.lower() == 'root':
                logger = root_logger
            else:
                logger = logging.getLogger(logger_class)
            configured_loggers[logger_class] = logger
            logger.propagate = True
            if desired_level_name is None:
                desired_level_name = 'INFO'
            else:
                desired_level_name = desired_level_name.upper()
            if self.trace_logging_setup:
                # Note: We can't use logging here yet
                print(f'Setting logger {logger.name} to {desired_level_name}')
            logger.setLevel(desired_level_name)

        # Will not include root logger
        for logger_class in sorted(logging.Logger.manager.loggerDict):
            logger = logging.getLogger(logger_class)
            if self.trace_logging_setup:
                print(f'Logger {logger_class} handlers {logger.handlers}')
            if logger_class not in configured_loggers:
                if self.trace_logging_setup:
                    # Note: We can't use logging here yet
                    print(
                        f"Checking existing logger {logger_class} "
                        f"level {logging.getLevelName(logger.getEffectiveLevel())}"
                    )
                for compare_logger in sorted(configured_loggers):
                    if logger_class.startswith(compare_logger):
                        parent_logger = configured_loggers[compare_logger]
                        level = parent_logger.getEffectiveLevel()
                        logger.setLevel(level)
                        if self.trace_logging_setup:
                            # Note: We can't use logging here yet
                            print(
                                f"Existing logger {logger_class} re-setup with {parent_logger.name} "
                                f"settings {logging.getLevelName(level)}"
                            )
                        logger.propagate = True

        if self.trace_logging_setup:
            # Note: We can't use logging here yet
            print(f"Root logging level is {logging.getLevelName(root_logger.getEffectiveLevel())}")
            print(f"Root logging handlers are {root_logger.handlers}")

    def setup_logging(
            self,
            log_file_prefix: str = None,
            add_date_to_log_file_name: bool = None,
            log_file_suffix: str = '.log',
            console_output=None,
            use_log_file_setting=True,
    ):
        """
        Setup logging based on configuration.
        """

        root_logger = logging.getLogger()

        if not hasattr(root_logger, 'config_wrangler_setup_done'):
            # Store info that we have already setup logging in the root logger.
            # This could be stored in the config, it would be too easy for methods to open
            # their own configs and thus will not know the logging had already been set up.
            root_logger.config_wrangler_setup_done = True

            # Close out any existing handlers
            for handler in root_logger.handlers:
                handler.flush()
                handler.close()

            # Reset the handlers
            root_logger.handlers.clear()

            if self.trace_logging_setup:
                # Note: We can't use logging here yet
                print('logging.trace_setup is True}')
                print(f'Starting root logger handlers {root_logger.handlers}')

            # Modify the root logger level to at least INFO for now. setup_log_levels could change it
            root_logger.setLevel(logging.INFO)

            if console_output is None:
                error_output = sys.stderr
                regular_output = sys.stdout
            else:
                error_output = console_output
                regular_output = console_output

            console_error_log = logging.StreamHandler(error_output)
            console_error_log.setLevel(logging.ERROR)
            root_logger.addHandler(console_error_log)

            console_log = logging.StreamHandler(regular_output)
            console_log.setLevel(self.console_log_level)

            # define a filter for non-error messages
            def non_error(record):
                return record.levelno != logging.ERROR

            # Errors go to console_error_log above, so we don't want them here as well
            console_log.addFilter(non_error)

            root_logger.addHandler(console_log)

            console_entry_format = self.console_entry_format
            if console_entry_format:
                console_entry_formatter = TZFormatter(console_entry_format, self.logging_date_format)
                console_log.setFormatter(console_entry_formatter)
                console_error_log.setFormatter(console_entry_formatter)

            self.setup_log_levels()

            # Switch to this modules logger
            log = logging.getLogger(__name__)

            logging.captureWarnings(True)

            log_level_name = logging.getLevelName(log.getEffectiveLevel())
            if self.trace_logging_setup:
                log.info(f"This modules logging level is {log_level_name}")

            if use_log_file_setting or log_file_prefix is not None:
                return self.add_log_file_handler(
                    log_file_prefix=log_file_prefix,
                    add_date_to_log_file_name=add_date_to_log_file_name,
                    log_file_suffix=log_file_suffix,
                )
            else:
                if self.trace_logging_setup:
                    log.info('use_log_file_setting = False. setup_log_file not called.')
                return None



import io
import logging
import warnings
from datetime import datetime, timezone
from enum import auto, Enum
from functools import lru_cache
from pathlib import PurePosixPath, Path, PurePath
from typing import *

from cachetools import cached, TTLCache
from pydantic import field_validator, PrivateAttr

from config_wrangler.config_exception import ConfigError
from config_wrangler.config_templates.aws.aws_session import AWS_Session

# noinspection PyProtectedMember

try:
    import boto3
    from boto3.s3.transfer import TransferConfig
    from botocore.exceptions import ClientError
    from botocore.response import StreamingBody

    # Unfortunately, the NoSuchKey exception below does not reliably work. ClientError is used instead
    # s3_unconnected_client = boto3.client('s3')
    # NoSuchKey = s3_unconnected_client.exceptions.NoSuchKey

    ERROR_S3_NOT_FOUND = {'404', 'NoSuchKey'}

    # Also note the following list of error codes:
    # https://docs.aws.amazon.com/AmazonS3/latest/API/ErrorResponses.html#ErrorCodeList
except ImportError:
    raise ImportError("S3_Bucket requires boto3 to be installed")

if TYPE_CHECKING:
    # https://youtype.github.io/boto3_stubs_docs/mypy_boto3_s3/client/
    from mypy_boto3_s3.client import S3Client

    # https://youtype.github.io/boto3_stubs_docs/mypy_boto3_s3/service_resource/
    from mypy_boto3_s3.service_resource import Bucket
    from mypy_boto3_s3.service_resource import Object
    from mypy_boto3_s3.service_resource import ObjectSummary
    from mypy_boto3_s3.service_resource import BucketObjectsCollection
    from mypy_boto3_s3.service_resource import S3ServiceResource

# NOTE: If you are not seeing boto3-stubs code completion in Intellij-based IDEs,
#       please increase the intellisense filesize limit
#       e.g `idea.max.intellisense.filesize=20000` in IDE custom properties
#       (Help > Edit Custom Properties), then restart.

local_timezone = datetime.now(timezone.utc).astimezone().tzinfo


class S3ClientError(ClientError):
    def __init__(self, message: str, original_error: Optional[ClientError]):
        Exception.__init__(self, message)
        if original_error is not None:
            self.response = original_error.response
            self.operation_name = original_error.operation_name
        else:
            self.response = {
                'Error': {
                    'Code': 'Unknown',
                    'Message': message,
                }
            }
            self.operation_name = 'Unknown'


# noinspection PyPep8Naming
class S3_Bucket(AWS_Session):
    bucket_name: str
    key: Optional[str] = None
    treat_as_folder: bool = False

    _service: str = PrivateAttr(default='s3')

    class OverwriteModes(Enum):
        ALWAYS_OVERWRITE = auto()
        OVERWRITE_OLDER = auto()
        NEVER_OVERWRITE = auto()

    def __str__(self):
        key = self._get_key()
        if key == '':
            return f"s3://{self.bucket_name}"
        else:
            return f"s3://{self.bucket_name}/{key}"

    def __hash__(self):
        return hash(str(self))

    @property
    def resource(self) -> 'S3ServiceResource':
        return super().resource

    @property
    def client(self) -> 'S3Client':
        return super().client

    def get_bucket_region_name(self) -> str:
        region = self.client.get_bucket_location(Bucket=self.bucket_name)['LocationConstraint']
        # Buckets in Region us-east-1 have a LocationConstraint of null
        if region is None:
            region = 'us-east-1'
        return region

    @staticmethod
    def _boto3_error(ex: ClientError) -> str:
        return ex.response.get('Error', {}).get('Code')

    @staticmethod
    def _boto3_error_match(ex: ClientError, error_set: Set[str]) -> bool:
        return S3_Bucket._boto3_error(ex) in error_set

    def get_boto3_bucket(self) -> 'Bucket':
        # Might raise error code NoSuchBucket
        return self.resource.Bucket(self.bucket_name)

    @staticmethod
    @lru_cache(maxsize=128)
    def _get_bucket_region(client, bucket_name) -> str:
        location = client.get_bucket_location(Bucket=bucket_name)
        if location is None or location['LocationConstraint'] is None:
            # https://docs.aws.amazon.com/AmazonS3/latest/API/API_CreateBucket.html#API_CreateBucket_RequestSyntax
            # If you don't specify a Region, the bucket is created in the US East (N. Virginia) Region (us-east-1).
            return 'us-east-1'
        else:
            return location['LocationConstraint']

    def get_bucket_region(self) -> str:
        """
        Get the region_name from the actual S3 bucket definition.

        NOTE:
            This can differ from the region_name attribute specified in the init call to this class
            or the config file that loads it.
            The region_name attribute is used for establishing the AWS session.
            get_bucket_region() is used to find out in which region the data is stored.
        """
        return S3_Bucket._get_bucket_region(self.client, self.bucket_name)

    @staticmethod
    def _non_blank_key(key: str):
        key_str = str(key).strip()
        if key is None or key_str in {''}:
            return False
        else:
            return True

    @staticmethod
    def _is_blank_key(key: str):
        return not S3_Bucket._non_blank_key(key)

    def _get_key(self, extra_key: Optional[Union[str, PurePosixPath]] = None) -> str:
        if self._is_blank_key(extra_key):
            if self._is_blank_key(self.key):
                key = ''
            else:
                key = self.key
        else:
            if self._is_blank_key(self.key):
                key = extra_key
            else:
                key = str(PurePosixPath(self.key) / extra_key)

        return key

    class CompareResult(Enum):
        DIFFERENT_SIZE = auto()
        LOCAL_NEWER = auto()
        SAME_TIMES = auto()
        LOCAL_OLDER = auto()

    @staticmethod
    def _compare_object_to_file(
            bucket_object: Union['Object', 'ObjectSummary'],
            local_filename: Path,
    ) -> CompareResult:
        s3_last_modified = bucket_object.last_modified
        s3_file_size = bucket_object.content_length

        local_stats = local_filename.stat()
        local_last_modified = datetime.fromtimestamp(
            local_stats.st_mtime,
            tz=local_timezone
        )
        local_file_size = local_stats.st_size

        if s3_file_size != local_file_size:
            return S3_Bucket.CompareResult.DIFFERENT_SIZE
        else:
            if local_last_modified > s3_last_modified:
                return S3_Bucket.CompareResult.LOCAL_NEWER
            elif local_last_modified == s3_last_modified:
                return S3_Bucket.CompareResult.SAME_TIMES
            else:
                return S3_Bucket.CompareResult.LOCAL_OLDER

    def upload_file(
            self,
            *,
            local_filename: Union[str, Path],
            key: Optional[Union[str, PurePosixPath]] = None,
            extra_args: Optional[dict] = None,
            transfer_config: Optional[TransferConfig] = None,
            overwrite_mode: OverwriteModes = OverwriteModes.ALWAYS_OVERWRITE
    ):
        if transfer_config is None:
            transfer_config = TransferConfig(multipart_threshold=5 * (1024**3))  # 5 GB

        if overwrite_mode == S3_Bucket.OverwriteModes.ALWAYS_OVERWRITE:
            do_upload = True
        else:
            try:
                bucket_object = self.get_object(key)
                if overwrite_mode == S3_Bucket.OverwriteModes.NEVER_OVERWRITE:
                    do_upload = False
                else:
                    local_filename = Path(local_filename)
                    compare_result = S3_Bucket._compare_object_to_file(bucket_object, local_filename)
                    if compare_result in {
                        S3_Bucket.CompareResult.DIFFERENT_SIZE,
                        S3_Bucket.CompareResult.LOCAL_NEWER
                    }:
                        do_upload = True
                    else:
                        do_upload = False
            except ClientError as ex:
                if self._boto3_error_match(ex, ERROR_S3_NOT_FOUND):
                    do_upload = True
                else:
                    raise

        if do_upload:
            key = self._get_key(key)
            self.client.upload_file(
                Filename=str(local_filename),
                Bucket=self.bucket_name,
                Key=key,
                ExtraArgs=extra_args,
                Config=transfer_config,
            )

    def open(
            self,
            mode: str = 'r',
            encoding: Optional[str] = None,
            errors: Optional[str] = None,
    ) -> io.IOBase:
        if mode is None or len(mode) == 0:
            raise ValueError('open mode required')
        if mode[0] != 'r':
            raise ValueError('S3 open mode only supports read (r, rt, rb)')

        as_text = True
        if len(mode) > 1:
            if mode[1] == 'b':
                as_text = False

        body: StreamingBody = self.get_object().get()['Body']
        # noinspection PyProtectedMember,PyTypeChecker
        bytes_stream = io.BufferedReader(body._raw_stream, 8192)

        if as_text:
            return io.TextIOWrapper(
                bytes_stream,
                encoding=encoding,
                errors=errors,
            )
        else:
            return bytes_stream

    def read_bytes(self) -> bytes:
        with self.open('rb') as f:
            return f.read()

    def read_text(self, encoding='utf-8', errors=None) -> str:
        with self.open('rb', encoding=encoding, errors=errors) as f:
            return f.read()

    def copy_to(
            self,
            target: Union[str, 'S3_Bucket_Key']
    ):
        if isinstance(target, str):
            target = self._build_s3_bucket_key(target)
        self_key = self._get_key()
        self.client.copy_object(
            Bucket=target.bucket_name,
            Key=target.key,
            CopySource=dict(
                Bucket=self.bucket_name,
                Key=self_key,
            )
        )

    def rename(self, target: str):
        self.copy_to(target)
        self.delete()

    def download_file(
            self,
            *,
            local_filename: Union[str, Path],
            key: Optional[Union[str, PurePosixPath]] = None,
            extra_args: Optional[dict] = None,
            transfer_config: Optional[TransferConfig] = None,
            create_parents: bool = True,
            overwrite_mode: OverwriteModes = OverwriteModes.OVERWRITE_OLDER,
            _bucket_object_summary: 'ObjectSummary' = None,
    ):
        if self._non_blank_key(key):
            file_obj = self / key
            file_obj.download_file(
                local_filename=local_filename,
                extra_args=extra_args,
                transfer_config=transfer_config,
                create_parents=create_parents,
                overwrite_mode=overwrite_mode,
                _bucket_object_summary=_bucket_object_summary,
            )
        else:
            log = logging.getLogger(f"{__name__}.download_file")
            local_path = Path(local_filename)
            if create_parents:
                local_path.parent.mkdir(parents=True, exist_ok=True)

            try:
                if overwrite_mode in {S3_Bucket.OverwriteModes.OVERWRITE_OLDER, S3_Bucket.OverwriteModes.NEVER_OVERWRITE}:
                    if local_path.exists():
                        if overwrite_mode == S3_Bucket.OverwriteModes.NEVER_OVERWRITE:
                            do_download = False
                            log.info(f"{self} download to {local_path} skipped since file exists and mode = {overwrite_mode}")
                        else:  # Check ages and size
                            if _bucket_object_summary is None:
                                bucket_object = self.get_object()
                            else:
                                bucket_object = _bucket_object_summary

                            compare_result = S3_Bucket._compare_object_to_file(bucket_object, local_path)

                            if compare_result in {
                                S3_Bucket.CompareResult.DIFFERENT_SIZE,
                                S3_Bucket.CompareResult.LOCAL_OLDER
                            }:
                                do_download = True
                                log.info(
                                    f"{self} download to {local_path} downloaded since file exists but {compare_result} "
                                    f"and mode = {overwrite_mode}"
                                )
                            else:
                                do_download = False
                                log.info(
                                    f"{self} download to {local_path} skipped since file exists {compare_result} "
                                    f"and mode = {overwrite_mode}"
                                )
                    else:
                        do_download = True
                        log.info(
                            f"{self} download to {local_path} downloaded since local file not not exist"
                        )
                else:
                    do_download = True
                    log.info(
                        f"{self} download to {local_path} downloaded since mode = {overwrite_mode}"
                    )
            except ClientError as ex:
                if self._boto3_error_match(ex, ERROR_S3_NOT_FOUND):
                    raise S3ClientError(f"{self} does not exist", ex)
                else:
                    raise S3ClientError(f"{self} with ID {self.user_id} get info yielded error {ex}", ex)
            except Exception as ex2:
                raise S3ClientError(f"{self} get info yielded error {ex2} {repr(ex2)}", None)

            if do_download:
                key = self._get_key()
                try:
                    self.resource.Bucket(self.bucket_name).download_file(
                        Key=key,
                        Filename=str(local_path),
                        ExtraArgs=extra_args,
                        Config=transfer_config,
                    )
                except ClientError as ex:
                    if self._boto3_error_match(ex, ERROR_S3_NOT_FOUND):
                        raise S3ClientError(f"{self} does not exist", ex)
                    else:
                        raise S3ClientError(f"{self} with ID {self.user_id} download yielded error {ex}", ex)

    def download_files(
            self,
            *,
            local_path: Union[str, Path],
            key: Optional[Union[str, PurePosixPath]] = None,
            treat_as_folder: bool = None,
            extra_args: Optional[dict] = None,
            transfer_config: Optional[TransferConfig] = None,
            create_parents: bool = True,
            overwrite_mode: OverwriteModes = OverwriteModes.OVERWRITE_OLDER,
    ) -> Iterable[Path]:
        if self._non_blank_key(key):
            file_obj = self / key
            return file_obj.download_files(
                local_path=local_path,
                extra_args=extra_args,
                transfer_config=transfer_config,
                create_parents=create_parents,
                overwrite_mode=overwrite_mode,
            )
        else:
            local_path = Path(local_path)
            if create_parents:
                local_path.parent.mkdir(parents=True, exist_ok=True)

            results = list()
            full_key = self._get_key()
            base_path = PurePosixPath(full_key)
            for s3_object in self.list_objects(treat_as_folder=treat_as_folder):
                s3_object_path = PurePosixPath(s3_object.key)
                try:
                    relative_key = s3_object_path.relative_to(base_path)
                except ValueError:
                    relative_key = s3_object_path.relative_to(base_path.parent)
                local_filename = local_path / relative_key
                s3_file = self._build_s3_bucket_key(s3_object.key)
                results.append(local_filename)
                s3_file.download_file(
                    local_filename=str(local_filename),
                    create_parents=create_parents,
                    extra_args=extra_args,
                    transfer_config=transfer_config,
                    overwrite_mode=overwrite_mode,
                    _bucket_object_summary=s3_object,
                )
            return results

    def exists(self, key: Optional[Union[str, PurePosixPath]] = None) -> bool:
        key = self._get_key(key)

        try:
            self.client.head_object(Bucket=self.bucket_name, Key=str(key))
            return True
        except ClientError as ex:
            if self._boto3_error_match(ex, ERROR_S3_NOT_FOUND):
                return False
            else:
                raise

    def key_exists(self, key: Union[str, PurePosixPath]) -> bool:
        warnings.warn(
            'The `key_exists` method is deprecated; use `exists` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        return self.exists(key)

    def get_object_uncached(self, key: Optional[Union[str, PurePosixPath]] = None) -> 'Object':
        key = self._get_key(key)
        return self.resource.Object(self.bucket_name, str(key))

    @cached(cache=TTLCache(maxsize=1024, ttl=10))
    def get_object(self, key: Optional[Union[str, PurePosixPath]] = None) -> 'Object':
        return self.get_object_uncached(key)

    def delete(
            self,
            key: Optional[Union[str, PurePosixPath]] = None,
            version_id: str = None,
    ):
        key = self._get_key(key)
        kwargs = {
            'Bucket': self.bucket_name,
            'Key': key,
        }
        if version_id is not None:
            kwargs['VersionId'] = version_id
        self.client.delete_object(**kwargs)

    def unlink(self, missing_ok=False):
        try:
            self.delete()
        except ClientError as ex:
            if self._boto3_error_match(ex, ERROR_S3_NOT_FOUND):
                if not missing_ok:
                    raise
            else:
                raise

    def delete_by_key(
            self,
            key: Union[str, PurePosixPath],
            version_id: str = None,
    ):
        warnings.warn(
            'The `delete_by_key` method is deprecated; use `delete` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        self.delete(key, version_id=version_id)

    def list_objects(
            self,
            key: Optional[Union[str, PurePosixPath]] = None,
            treat_as_folder: bool = None,
    ) -> 'BucketObjectsCollection':
        log = logging.getLogger(f"{__name__}.list_objects")
        if treat_as_folder is None and key is not None:
            # If directly provided a key string, use it as is
            treat_as_folder = False
            log.debug(f"Using provided key {key}, defaulting treat_as_folder = False")
        key = self._get_key(key)
        if treat_as_folder is None:
            log.debug(f"Defaulting treat_as_folder to setting value {self.treat_as_folder}")
            treat_as_folder = self.treat_as_folder
        if treat_as_folder:
            if key != '' and not key.endswith('/'):
                key = f"{key}/"
                log.debug(f"treat_as_folder True, adding / to the end = {key}")
        try:
            collection = self.resource.Bucket(self.bucket_name).objects.filter(Prefix=key)
        except ClientError as ex:
            if self._boto3_error_match(ex, ERROR_S3_NOT_FOUND):
                raise S3ClientError(f"{self} with key={key} does not exist", ex)
            else:
                raise S3ClientError(f"{self} with key={key} list_objects yielded error {ex}", ex)
        return collection

    def find_objects(self, key: Union[str, PurePosixPath] = None) -> 'BucketObjectsCollection':
        warnings.warn(
            'The `find_objects` method is deprecated; use `list_objects` instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        return self.list_objects(key=key)

    def list_object_keys(
            self,
            key: Optional[Union[str, PurePosixPath]] = None,
            treat_as_folder: bool = None,
    ) -> List[str]:
        obj_collection = self.list_objects(key,treat_as_folder=treat_as_folder)
        return [obj.key for obj in obj_collection]

    def list_object_paths(
            self,
            key: Optional[Union[str, PurePosixPath]] = None,
            treat_as_folder: bool = None,
    ) -> List[PurePosixPath]:
        """
        Return the relative paths of objects contained in the in/under this object
        or, if provided, under the object + provided key parameter.


        """
        resolved_key = self._get_key(key)
        return [
            PurePosixPath(obj_key).relative_to(resolved_key)
            for obj_key in self.list_object_keys(key, treat_as_folder=treat_as_folder)
        ]

    # noinspection SpellCheckingInspection
    def iterdir(self) -> Iterable['S3_Bucket_Key']:
        """
        Return the S3_Bucket_Key objects contained in the in/under this object.
        """
        return [self / key for key in self.list_object_paths(treat_as_folder=True)]

    @staticmethod
    def _path_to_key(path: PurePosixPath):
        if path.name == '' and path.parent == PurePosixPath('.'):
            return ''
        else:
            return str(path)

    @property
    def parent(self):
        key = PurePosixPath(self._get_key()).parent
        return self._factory(S3_Bucket_Key, key=self._path_to_key(key))

    @property
    def parents(self):
        return [self._factory(S3_Bucket_Key, key=self._path_to_key(key))
                for key in PurePosixPath(self._get_key()).parents
                ]

    @property
    def parts(self):
        key = PurePosixPath(self._get_key())
        return [self.bucket_name].extend(key.parts)

    @property
    def name(self):
        key = PurePosixPath(self._get_key())
        return key.name

    @property
    def suffix(self):
        key = PurePosixPath(self._get_key())
        return key.suffix

    @property
    def suffixes(self):
        key = PurePosixPath(self._get_key())
        return key.suffixes

    @property
    def stem(self):
        key = PurePosixPath(self._get_key())
        return key.stem

    def is_relative_to(self, other: 'S3_Bucket'):
        if self.bucket_name != other.bucket_name:
            return False
        else:
            key = PurePosixPath(self._get_key())
            other_key = PurePosixPath(other._get_key())
            return key.relative_to(other_key)

    def with_name(self, name: str):
        key = PurePosixPath(self._get_key())
        return self._build_s3_bucket_key(str(key.with_name(name)))

    def with_stem(self, stem: str):
        key = PurePosixPath(self._get_key())
        return self._build_s3_bucket_key(str(key.with_stem(stem)))

    def with_suffix(self, suffix: str):
        key = PurePosixPath(self._get_key())
        return self._build_s3_bucket_key(str(key.with_suffix(suffix)))

    def content_type(self) -> str:
        s3_obj = self.get_object()
        val = s3_obj.get()['ContentType']
        return val

    def is_file(self):
        try:
            return self.content_type() != 'application/x-directory'
        except ClientError as ex:
            if self._boto3_error_match(ex, ERROR_S3_NOT_FOUND):
                return False
            raise

    # Note: There is no really useful & consistent way to implement is_dir on S3
    # A good investigation related to this is here:
    # https://www.tecracer.com/blog/2023/01/what-are-the-folders-in-the-s3-console.html

    def __truediv__(
            self, other: Union[str, PurePath]
    ) -> 'S3_Bucket_Key':
        key = self._get_key(other)
        return self._build_s3_bucket_key(key, treat_as_folder=True)

    # noinspection SpellCheckingInspection
    def joinpath(
        self,
        *others
    ) -> Union[
        'S3_Bucket_Key',
        'S3_Bucket_Folder'
    ]:
        new_path = self
        for other in others:
            new_path = new_path / other
        return new_path

    def _build_s3_bucket_folder(self, folder: Union[str, Path]):
        return self._factory(
            S3_Bucket_Folder,
            exclude={'file_name'},
            folder=str(folder)
        )

    def _build_s3_bucket_folder_file(self, file_name: Union[str, Path], folder: Union[str, Path] = None):
        if folder is None:
            return self._factory(
                S3_Bucket_Folder_File,
                exclude={'key'},
                file_name=str(file_name)
            )
        else:
            return self._factory(
                S3_Bucket_Folder_File,
                exclude={'key'},
                folder=str(folder),
                file_name=str(file_name)
            )

    def _build_s3_bucket_key(
            self,
            key: Union[str, Path],
            treat_as_folder: bool = None,
    ):
        s3_key_obj =  self._factory(S3_Bucket_Key, key=str(key), exclude={'folder', 'file_name'})
        if treat_as_folder is not None:
            s3_key_obj.treat_as_folder = treat_as_folder
        return s3_key_obj


# noinspection PyPep8Naming
class S3_Bucket_Key(S3_Bucket):
    """
    Represents a unique file (key) within an S3 bucket.
    Similar to S3_Bucket but the key is required
    """
    key: str

    # Note the order of decorators matters!
    # noinspection PyMethodParameters,PyNestedDecorators
    @field_validator('key')
    @classmethod
    def validate_key(cls, v):
        # Handle pathlib values
        if not isinstance(v, str):
            v = str(v)
        if len(v) == 0:
            raise ConfigError(f"Zero length string not a valid key")
        return v

    def upload_specified_file(
            self,
            local_filename: Union[str, PurePosixPath],
            extra_args: Optional[dict] = None,
            transfer_config: Optional[TransferConfig] = None,
    ):
        warnings.warn(
            "The `upload_folder_file` method is deprecated; use `my_bucket_key.upload_file() instead",
            DeprecationWarning,
            stacklevel=2,
        )
        super().upload_file(
            local_filename=local_filename,
            extra_args=extra_args,
            transfer_config=transfer_config,
        )

    def download_specified_file(
            self,
            local_filename: Union[str, PurePosixPath],
            extra_args: Optional[dict] = None,
            transfer_config: Optional[TransferConfig] = None,
            create_parents: bool = True,
    ):
        warnings.warn(
            "The `download_specified_file` method is deprecated; use `my_bucket_key.download_file() instead",
            DeprecationWarning,
            stacklevel=2,
        )
        super().download_file(
            local_filename=local_filename,
            extra_args=extra_args,
            transfer_config=transfer_config,
            create_parents=create_parents,
        )


# noinspection PyPep8Naming
class S3_Bucket_Folder(S3_Bucket):
    """
        Represents a folder within an S3 bucket.
    """
    folder: str
    treat_as_folder: bool = True

    # Note the order of decorators matters!
    # noinspection PyMethodParameters,PyNestedDecorators
    @field_validator('folder')
    @classmethod
    def validate_folder(cls, v):
        # Handle pathlib values
        if not isinstance(v, str):
            v = str(v)
        if len(v) == 0:
            raise ValueError(f"Zero length string not a valid folder")
        return v

    # TODO: A folder and a key are not the same.  Folders should always send a trailing slash to S3

    def _get_key(self, extra_key: Optional[Union[str, PurePosixPath]] = None) -> str:
        if self._non_blank_key(extra_key):
            if self.folder == '/':
                return extra_key
            else:
                return str(PurePosixPath(self.folder) / extra_key)
        else:
            return self.folder

    def upload_folder_file(
            self,
            local_filename: Union[str, Path],
            key_suffix: Union[str, PurePosixPath],
            extra_args: Optional[dict] = None,
            transfer_config: Optional[TransferConfig] = None,
    ):
        """
        Differs from upload_file in that it requires a key_suffix to add after the folder name
        """
        warnings.warn(
            "The `upload_folder_file` method is deprecated; use `(my_folder / key_suffix).download_file() instead",
            DeprecationWarning,
            stacklevel=2,
        )

        full_key = PurePosixPath(self.folder, key_suffix)
        super().upload_file(
            local_filename=local_filename,
            key=full_key,
            extra_args=extra_args,
            transfer_config=transfer_config,
        )

    def download_folder_file(
            self,
            key_suffix: Union[str, PurePosixPath],
            local_filename: Union[str, Path],
            extra_args: Optional[dict] = None,
            transfer_config: Optional[TransferConfig] = None,
            create_parents: bool = True,
    ):
        """
        Differs from download_file in that it requires a key_suffix to add after the folder name
        """
        warnings.warn(
            "The `download_folder_file` method is deprecated; use `(my_folder / key_suffix).upload_file() instead",
            DeprecationWarning,
            stacklevel=2,
        )

        full_key = PurePosixPath(self.folder, key_suffix)
        super().download_file(
            key=full_key,
            local_filename=local_filename,
            extra_args=extra_args,
            transfer_config=transfer_config,
            create_parents=create_parents,
        )


# noinspection PyPep8Naming
class S3_Bucket_Folder_File(S3_Bucket_Folder):
    """
        Represents a unique folder & file within an S3 bucket.
        Similar to S3_Bucket_Key but uses folder + file_name instead of a single key.
    """
    file_name: str
    treat_as_folder: bool = False

    # Note the order of decorators matters!
    # noinspection PyNestedDecorators
    @field_validator('file_name')
    @classmethod
    def validate_file_name(cls, v):
        # Handle pathlib values
        if not isinstance(v, str):
            v = str(v)
        if len(v) == 0:
            raise ConfigError(f"Zero length string not a valid file_name")
        return v

    def _get_key(self, extra_key: Optional[Union[str, PurePosixPath]] = None) -> str:
        if self._non_blank_key(extra_key):
            return str(PurePosixPath(self.folder) / self.file_name / extra_key)
        else:
            return str(PurePosixPath(self.folder) / self.file_name)

    def upload_specified_file(
        self,
        local_filename: Union[str, PurePosixPath],
        extra_args: Optional[dict] = None,
        transfer_config: Optional[TransferConfig] = None,
    ):
        warnings.warn(
            "The `upload_folder_file` method is deprecated; use `my_folder_file.upload_file() instead",
            DeprecationWarning,
            stacklevel=2,
        )

        super().upload_file(
            local_filename=local_filename,
            extra_args=extra_args,
            transfer_config=transfer_config,
        )

    def download_specified_file(
        self,
        local_filename: Union[str, PurePosixPath],
        extra_args: Optional[dict] = None,
        transfer_config: Optional[TransferConfig] = None,
    ):
        warnings.warn(
            "The `download_specified_file` method is deprecated; use `my_folder_file.download_file() instead",
            DeprecationWarning,
            stacklevel=2,
        )

        super().download_file(
            local_filename=local_filename,
            extra_args=extra_args,
            transfer_config=transfer_config,
        )



from typing import TYPE_CHECKING, Union


from pydantic import PrivateAttr
from config_wrangler.config_templates.aws.aws_session import AWS_Session

try:
    import boto3
    from botocore.exceptions import ClientError
except ImportError:
    raise ImportError("Lambda requires boto3 to be installed")

if TYPE_CHECKING:
    # https://youtype.github.io/boto3_stubs_docs/mypy_boto3_s3/client/
    from mypy_boto3_lambda import LambdaClient
    from mypy_boto3_lambda.type_defs import InvocationResponseTypeDef, InvokeWithResponseStreamResponseTypeDef


class Lambda(AWS_Session):
    _service: str = PrivateAttr(default='lambda')

    @property
    def client(self) -> 'LambdaClient':
        return super().client

    def invoke_function(
            self,
            function_name: str,
            invocation_type: str = 'RequestResponse',
            qualifier: str = ...,
            payload: Union[str, bytes] = ...,
            client_context: str = ...,
            include_log_tail: bool = False,
    ) -> 'InvocationResponseTypeDef':
        if isinstance(payload, str):
            payload = payload.encode('utf-8')

        kwargs = {}
        if qualifier is not ...:
            kwargs['Qualifier'] = qualifier
        if payload is not ...:
            kwargs['Payload'] = payload
        if include_log_tail:
            kwargs['LogType'] = 'Tail'
        if client_context is not ...:
            kwargs['ClientContext'] = client_context

        # noinspection PyTypeChecker
        return self.client.invoke(
            FunctionName=function_name,
            InvocationType=invocation_type,
            **kwargs
        )

    def invoke_with_response_stream(
            self,
            function_name: str,
            invocation_type: str = 'RequestResponse',
            qualifier: str = ...,
            payload: Union[str, bytes] = ...,
            client_context: str = ...,
            include_log_tail: bool = False,
            ) -> 'InvokeWithResponseStreamResponseTypeDef':
        if isinstance(payload, str):
            payload = payload.encode('utf-8')

        kwargs = {}
        if qualifier is not ...:
            kwargs['Qualifier'] = qualifier
        if payload is not ...:
            kwargs['Payload'] = payload
        if include_log_tail:
            kwargs['LogType'] = 'Tail'
        if client_context is not ...:
            kwargs['ClientContext'] = client_context

        # noinspection PyTypeChecker
        return self.client.invoke_with_response_stream(
            FunctionName=function_name,
            InvocationType=invocation_type,
            **kwargs
        )



from datetime import date, datetime, time
from enum import Enum
from pathlib import Path
from typing import *

from pydantic import BaseModel

from config_wrangler.config_data_loaders.file_config_data_loader import FileConfigDataLoader
from config_wrangler.config_types.dynamically_referenced import ListDynamicallyReferenced
from config_wrangler.utils import lenient_issubclass


class TomlConfigDataLoader(FileConfigDataLoader):
    def __init__(
            self,
            file_name: str,
            start_path: Optional[str] = None,
    ):
        super().__init__(
            start_path=start_path,
            file_name=file_name,
        )
        try:
            import toml
        except ImportError:
            raise RuntimeError(f"Module toml required for TomlSettingsLoader. "
                               f"Use pip install toml or poetry add toml as appropriate.")
        self.toml = toml

    def _read_file(self, file_path: Path) -> MutableMapping:
        self.log.info(f"Reading {file_path}")
        self.files_read.append(file_path)
        with file_path.open('rt', encoding='utf8') as toml_content:
            config_data = self.toml.load(toml_content)
        return config_data

    def save_config_data(self, config_data: BaseModel):
        file_path = Path(self.start_path, self.file_name)
        config_data_toml_ready = TomlConfigDataLoader.prepare_config_data_for_save(config_data)
        with file_path.open('wt', encoding='utf8') as config_file:
            config_file.write(self.toml.dumps(config_data_toml_ready))
        self.log.info(f"Created {file_path}")

    @staticmethod
    def format_value_for_save(field_value):
        if isinstance(field_value, bool):
            pass
        elif isinstance(field_value, int):
            pass
        elif isinstance(field_value, float):
            pass
        elif isinstance(field_value, datetime):
            pass
        elif isinstance(field_value, date):
            pass
        elif isinstance(field_value, time):
            pass
        elif isinstance(field_value, bytes):
            field_value = field_value.decode('utf8')
        elif isinstance(field_value, Path):
            field_value = str(field_value)
        elif isinstance(field_value, Enum):
            field_value = str(field_value)
        elif isinstance(field_value, dict):
            str_keys = True
            for key, value in field_value.items():
                if not isinstance(key, str):
                    str_keys = False
                field_value[key] = TomlConfigDataLoader.format_value_for_save(value)
            if not str_keys:
                field_value = str(field_value)
            else:
                pass
        else:
            field_value = str(field_value)
        return field_value

    @staticmethod
    def prepare_config_data_for_save(config: BaseModel, default_delimiter='\n', parents=None) -> dict:
        if parents is None:
            parents = []
        config_data_dict = config.model_dump()
        for field_name, field_info in config.model_fields.items():
            field_name = field_info.alias or field_name
            field_value = config_data_dict[field_name]
            if lenient_issubclass(field_info.annotation, BaseModel):
                config_data_dict[field_name] = TomlConfigDataLoader.prepare_config_data_for_save(
                    getattr(config, field_name),
                    parents=parents + [field_name]
                )
            elif lenient_issubclass(field_info.annotation, ListDynamicallyReferenced):
                section_name_list = []
                for sub_section_number, sub_section_value in enumerate(getattr(config, field_name)):
                    sub_section_id = f"{field_name}_{sub_section_number}"
                    section_name_list.append(sub_section_id)
                    config_data_dict[sub_section_id] = TomlConfigDataLoader.prepare_config_data_for_save(
                        sub_section_value,
                    )
                config_data_dict[field_name] = section_name_list
            elif lenient_issubclass(field_info.annotation, List):
                value_list = [TomlConfigDataLoader.format_value_for_save(v) for v in field_value]
                config_data_dict[field_name] = value_list
            elif lenient_issubclass(field_info.annotation, Set):
                value_list = [TomlConfigDataLoader.format_value_for_save(v) for v in field_value]
                config_data_dict[field_name] = value_list
            else:
                # Use python format
                config_data_dict[field_name] = TomlConfigDataLoader.format_value_for_save(field_value)

        return config_data_dict



from typing import TYPE_CHECKING, List

from pydantic import PrivateAttr

from config_wrangler.config_templates.aws.aws_session import AWS_Session

if TYPE_CHECKING:
    # https://youtype.github.io/boto3_stubs_docs/mypy_boto3_ssm/
    from mypy_boto3_ssm.client import SSMClient
    from mypy_boto3_ssm.type_defs import (
        GetParameterResultTypeDef,
        GetParametersResultTypeDef,
        GetParametersByPathResultTypeDef,
    )


class SSM(AWS_Session):
    _service: str = PrivateAttr(default='ssm')
    # Region is not optional for SSM
    region_name: str

    @property
    def client(self) -> 'SSMClient':
        return super().client

    def get_parameters_response(
            self,
            names: List[str],
            with_decryption: bool = True,
    ) -> 'GetParametersResultTypeDef':
        return self.client.get_parameters(
            Names=names,
            WithDecryption=with_decryption,
        )

    def get_parameters(
            self,
            names: List[str],
            with_decryption: bool = True,
    ) -> dict:
        response = self.get_parameters_response(names=names, with_decryption=with_decryption)
        return {entry['Name']: entry['Value'] for entry in response['Parameters']}

    def get_parameters_by_path_response(
            self,
            path: str,
            recursive: bool = True,
            with_decryption: bool = True,
    ) -> 'GetParametersByPathResultTypeDef':
        return self.client.get_parameters_by_path(
            Path=path,
            Recursive=recursive,
            WithDecryption=with_decryption,
        )

    def get_parameters_by_path(
            self,
            path: str,
            recursive: bool = True,
            with_decryption: bool = True,
    ) -> dict:
        response = self.get_parameters_by_path_response(
            path=path,
            recursive=recursive,
            with_decryption=with_decryption,
        )
        return {
            entry['Name']: entry['Value']
            for entry in response['Parameters']
        }

    def get_parameter_response(
            self,
            name: str,
            with_decryption: bool = True,
    ) -> 'GetParameterResultTypeDef':
        return self.client.get_parameter(
            Name=name,
            WithDecryption=with_decryption,
        )

    def get_parameter_value(
            self,
            name: str,
            with_decryption: bool = True,
    ) -> str:
        ssm_response = self.get_parameter_response(name, with_decryption)
        return ssm_response['Parameter']['Value']


# noinspection PyPep8Naming
class SSM_Parameter(SSM):
    parameter_name: str

    def get_value(self):
        return self.get_parameter_value(self.parameter_name)



import ast
import json
import logging
import re
import types
import warnings
from datetime import timezone, datetime
from enum import Enum, auto
from typing import *

from pydantic import BaseModel, ValidationError
from pydantic.fields import FieldInfo
from pydicti import dicti, Dicti

from config_wrangler.config_exception import ConfigError
from config_wrangler.config_types.delimited_field import DelimitedListFieldInfo
from config_wrangler.config_types.dynamically_referenced import DynamicallyReferenced, DynamicFieldInfo


# Moved here because  Pydantic V2 deprecated it
def lenient_issubclass(
        cls: Any,
        class_or_tuple: Union[Type[Any], Tuple[Type[Any], ...], Set[Type[Any]], None]
) -> bool:
    try:
        if isinstance(class_or_tuple, set):
            class_or_tuple = tuple(class_or_tuple)

        if isinstance(cls, type):
            return issubclass(cls, class_or_tuple)
        else:
            try:
                origin = cls.__origin__
            except AttributeError:
                origin = get_origin(cls)
            if origin is not None:
                if origin == Union:
                    for union_cls in get_args(cls):
                        if lenient_issubclass(union_cls, class_or_tuple):
                            return True
                    return False
                else:
                    return issubclass(origin, class_or_tuple)

    except TypeError:
        if isinstance(cls, (types.GenericAlias, types.UnionType)):
            return False
        raise


def get_inner_type(cls: Any):
    try:
        origin = cls.__origin__
    except AttributeError:
        origin = get_origin(cls)

    if origin == Union:
        for inner_cls in get_args(cls):
            return get_inner_type(inner_cls)
    else:
        return get_args(cls)


def has_sub_fields(inner_type: Type):
    return hasattr(inner_type, 'model_fields')


class TZFormatter(logging.Formatter):
    local_timezone = datetime.now(timezone.utc).astimezone().tzinfo

    @staticmethod
    def tz_aware_converter(timestamp) -> datetime:
        return datetime.fromtimestamp(timestamp, TZFormatter.local_timezone)

    def formatTime(self, record, datefmt=None):
        dt = TZFormatter.tz_aware_converter(record.created)
        if datefmt is not None:
            result = dt.strftime(datefmt)
        else:
            time_part = dt.strftime(self.default_time_format)
            result = self.default_msec_format % (time_part, record.msecs)
        return result


def merge_configs(child: MutableMapping, parent: MutableMapping) -> None:
    for section in parent:
        if section not in child:
            child[section] = parent[section]
        else:
            if isinstance(child[section], MutableMapping):
                merge_configs(child[section], parent[section])
            else:
                pass


def resolve_variable(root_config_data: MutableMapping, variable_name: str, part_delimiter=':') -> Any:
    variable_name_parts = variable_name.split(part_delimiter)
    result = root_config_data
    for part in variable_name_parts:
        # Change to case-insensitive dict
        if not isinstance(result, dicti):
            result = Dicti(result)

        if part in result:
            result = result[part]
        else:
            raise ValueError(f"<<{part} NOT FOUND when resolving variable with parts: {variable_name_parts}>>")
    return result


_interpolation_re = re.compile(r"\${([^}]+)}")

def interpolate_value(*, value: str, container: MutableMapping, root_config_data: MutableMapping) -> str:
    """
    Throws: ValueError if value can not be interpolated
    """
    if '$' not in value:
        return value
    else:
        depth = 0
        done = False
        new_value = value
        while not done:
            done = True
            depth += 1
            variables_cnt = 0
            result_values = []
            next_start = 0
            if isinstance(new_value, str):
                for variable_found in _interpolation_re.finditer(new_value):
                    variables_cnt += 1
                    variable_name = variable_found.groups()[0]
                    var_start, var_end = variable_found.span()

                    part_delimiter = None
                    if ':' in variable_name:
                        part_delimiter = ':'

                    elif '.' in variable_name:
                        part_delimiter = '.'

                    variable_replacement = 'ERROR'
                    if part_delimiter is not None:
                        try:
                            variable_replacement = resolve_variable(
                                root_config_data,
                                variable_name,
                                part_delimiter=part_delimiter,
                            )
                        except ValueError as e1:
                            try:
                                variable_replacement = resolve_variable(
                                    container,
                                    variable_name,
                                    part_delimiter=part_delimiter,
                                )
                            except ValueError as e2:
                                ValueError(f"<<{e1} resolving {variable_name}>>")
                    else:
                        try:
                            # Search in the local container instead of the root
                            # Change to case-insensitive dict
                            search_container = Dicti(container)
                            variable_replacement = search_container[variable_name]
                        except KeyError:
                            raise ValueError(f"<<{variable_name} NOT FOUND>>",)

                    result_values.append(new_value[next_start:var_start])
                    result_values.append(variable_replacement)
                    next_start = var_end
                if variables_cnt > 0:
                    if next_start < len(new_value):
                        result_values.append(new_value[next_start:])
                    result_values = [part for part in result_values if part != '']
                    if len(result_values) == 1:
                        # Possibly not a string -- maybe a dict
                        new_value = result_values[0]
                        if not isinstance(new_value, str):
                            done = True
                    else:
                        new_value = ''.join([str(v) for v in result_values])

                    if depth < 50:
                        done = False
                    else:
                        raise ValueError(
                            f"Interpolation recursion depth limit reached on value {value} "
                            f"ended processing with {new_value}"
                        )
        return new_value

class ContainerType(Enum):
    Mapping = auto()
    List = auto()
    Tuple = auto()
    Model = auto()

def set_container_value(
        container_type: ContainerType,
        container: Union[MutableMapping,List,BaseModel],
        attr: Union[str, int],
        value: Any
):
    if container_type == ContainerType.Mapping:
        container[attr] = value
    elif container_type == ContainerType.List:
        container[attr] = value
    elif container_type == ContainerType.Tuple:
        raise ValueError(
            f"Can't interpolate {value} inside tuple. Make it a list! See {breadcrumbs}"
        )
    else: # Model
        setattr(container, attr, value)

def interpolate_values(
        container: Union[MutableMapping,List,BaseModel],
        root_config_data: MutableMapping,
        breadcrumbs: List[str] = None,
) -> List[Tuple[str, str]]:
    errors = []
    if breadcrumbs is None:
        breadcrumbs = []
    if isinstance(container, MutableMapping):
        container_type = ContainerType.Mapping
        value_tuples = container.items()
    elif isinstance(container, list):
        container_type = ContainerType.List
        value_tuples = list(enumerate(container))
    elif isinstance(container, tuple):
        # We can't update the tuple so interpolation won't work
        # Howver, caller should change it to a list
        raise ValueError(
            f"Can't interpolate {value} inside tuple to {new_value}. Make it a list! See {breadcrumbs}"
        )
    else:
        container_type = ContainerType.Model
        value_tuples = list(container)

    for attr, value in value_tuples:
        if isinstance(value, MutableMapping) or isinstance(value, BaseModel) or isinstance(value, list) or isinstance(value, tuple):
            if isinstance(value, tuple):
                warnings.warn(f"Value of {attr} changed from tuple to list")
                # Convert to a list
                value = list(value)
                set_container_value(
                    container_type=container_type,
                    container=container,
                    attr=attr,
                    value=value,
                )
            sub_errors = interpolate_values(container=value, root_config_data=root_config_data, breadcrumbs=breadcrumbs + [attr])
            errors.extend(sub_errors)
        elif isinstance(value, str):
            try:
                new_value = interpolate_value(value=value, container=container, root_config_data=root_config_data)
                if new_value != value:
                    set_container_value(
                        container_type=container_type,
                        container=container,
                        value=new_value,
                        attr=attr,
                    )
            except ValueError as e:
                errors.append(('.'.join(breadcrumbs), str(e)))
    return errors

def parse_as_literal_or_json(value: str) -> Any:
    try:
        return ast.literal_eval(value)
    except (ValueError, SyntaxError) as el:
        try:
            return json.loads(value)
        except ValueError as ej:
            raise ConfigError(
                f"Value {value} could not be parsed as python literal '{el}' or json '{ej}'"
            )


def parse_delimited_list(
    field_name: str,
    field_info: FieldInfo,
    field_value: str
) -> Sequence:
    value = field_value.strip()
    if len(value) == 0:
        return []

    if isinstance(field_info, DelimitedListFieldInfo):
        delimiter = field_info.delimiter
    else:
        delimiter = None

    if delimiter is None and value[0] not in {'[', '{'}:
        # Try to automatically recognize the delimiter
        for try_delimiter in ['\n', ',', '|']:
            if try_delimiter in value:
                delimiter = try_delimiter
                break
    if delimiter is not None:
        if value[0] == delimiter:
            value = value[1:]
        result = [v.strip() for v in value.split(delimiter)]
        if lenient_issubclass(field_info.annotation, int):
            result = [int(v) for v in result]
        elif lenient_issubclass(field_info.annotation, float):
            result = [float(v) for v in result]
    else:
        try:
            result = parse_as_literal_or_json(value)
        except ValueError as e:
            # Single value list
            result = [value]
    return result


def full_name(parents: List[str], field_name: str):
    return '.'.join(parents + [field_name])


def inherit_fill(parent_config, child_config):
    for inherit_key, inherit_value in parent_config.items():
        if inherit_key not in child_config:
            child_config[inherit_key] = inherit_value


def find_referenced_section(
        field_name: str,
        field_info: FieldInfo,
        parents: List[str],
        section_name: Union[str, MutableMapping],
        current_dict: MutableMapping,
        root_dict: MutableMapping
) -> MutableMapping:
    try:
        # noinspection PyUnresolvedReferences
        inherit = field_info.inherit
    except AttributeError:
        inherit = False
    if isinstance(section_name, str):
        section_value = dict()
        section_name_parts = section_name.split('.')
        parts_used = []
        for parts_to_use in range(len(section_name_parts), 0, -1):
            section_name_2 = '.'.join(section_name_parts[:parts_to_use])
            if section_name_2 in root_dict:
                parts_used.append(tuple(section_name_2,))
                inherit_fill(root_dict[section_name_2], section_value)
            elif section_name_2 in current_dict:
                parts_used.append((*parents, section_name_2,))
                inherit_fill(current_dict[section_name_2], section_value)
        if inherit:
            inherit_fill(current_dict, section_value)
        if len(parts_used) == 0:
            raise ConfigError(
                f"{full_name(parents, field_name)} refers to section {section_name} which does not exist."
            )
        return section_value
    else:
        return section_name


def build_referenced_objects(
    field_name: str,
    field_info: FieldInfo,
    parents: List[str],
    parent_container: MutableMapping,
    root_config_data: MutableMapping,
    list_of_sections: Sequence[str],
    inner_type: type[BaseModel],
) -> Dict[str, BaseModel]:
    section_name = None
    try:
        section_contents = dict()
        for section_name in list_of_sections:
            # For DynamicallyReferenced we don't use the full value returned
            # here, but this helps us by validating the reference
            ref_section_value = find_referenced_section(
                field_name=field_name,
                field_info=field_info,
                parents=parents,
                section_name=section_name,
                current_dict=parent_container,
                root_dict=root_config_data,
            )
            if issubclass(inner_type, DynamicallyReferenced):
                # In the case of DynamicallyReferenced it refers to an existing static instance,
                # so we don't build a new instance, we instead just store the pointer.
                inner_type_instance = DynamicallyReferenced(ref=section_name)
            else:
                inner_type_instance = inner_type(**ref_section_value)
            section_contents[section_name] = inner_type_instance
        return section_contents
    except ValidationError as e:
        raise ValueError(f"Field {full_name(parents, field_name)} error on section {section_name} = {repr(e)})")


def match_config_data_to_field(
        field_name: str,
        field_info: FieldInfo,
        field_value: object,
        parent_container: MutableMapping,
        root_config_data: MutableMapping,
        parents: List[str],
):
    if lenient_issubclass(field_info.annotation, (str, int, float)):
        pass
    elif lenient_issubclass(field_info.annotation, list):
        if isinstance(field_value, str):
            field_value = parse_delimited_list(field_name, field_info, field_value)

        inner_type_args = get_inner_type(field_info.annotation)
        if len(inner_type_args) > 1:
            raise SyntaxError(
                f"{full_name(parents, field_name)} has type {field_info.annotation} "
                f"with more than one inner type {inner_type_args}"
            )
        elif len(inner_type_args) == 1:
            inner_type = inner_type_args[0]
            if has_sub_fields(inner_type):
                ref_object_dict = build_referenced_objects(
                    field_name=field_name,
                    field_info=field_info,
                    parents=parents,
                    parent_container=parent_container,
                    root_config_data=root_config_data,
                    list_of_sections=field_value,
                    inner_type=inner_type,
                )
                field_value = list(ref_object_dict.values())
    elif lenient_issubclass(field_info.annotation, tuple):
        if isinstance(field_value, str):
            field_value = parse_delimited_list(field_name, field_info, field_value)

        # See docs for Tuple annotations
        # https://docs.python.org/3/library/typing.html#annotating-tuples
        inner_type_args = get_inner_type(field_info.annotation)
        if len(inner_type_args) == 0:
            # Nothing to do for untyped tuple container
            pass
        elif len(inner_type_args) == 2 and inner_type_args[1] == Ellipsis:
            inner_type = inner_type_args[0]
            if has_sub_fields(inner_type):
                ref_object_dict = build_referenced_objects(
                    field_name=field_name,
                    field_info=field_info,
                    parents=parents,
                    parent_container=parent_container,
                    root_config_data=root_config_data,
                    list_of_sections=field_value,
                    inner_type=inner_type,
                )
                field_value = list(ref_object_dict.values())
        elif len(inner_type_args) != len(field_value):
            raise ValueError(
                f"{full_name(parents, field_name)} has type {field_info.annotation} "
                f"expects {len(inner_type_args)} values but got {len(field_value)} values."
            )
        else:
            new_field_values = list()
            for inner_type, value in zip(inner_type_args, field_value):
                if has_sub_fields(inner_type):
                    ref_object_dict = build_referenced_objects(
                        field_name=field_name,
                        field_info=field_info,
                        parents=parents,
                        parent_container=parent_container,
                        root_config_data=root_config_data,
                        list_of_sections=[value],
                        inner_type=inner_type,
                    )
                    new_field_values.append(ref_object_dict[value])
                else:
                    new_field_values.append(value)
    elif lenient_issubclass(field_info.annotation, dict):
        if isinstance(field_value, str):
            try:
                field_value = parse_as_literal_or_json(field_value)
            except ValueError as e:
                try:
                    list_of_sections = parse_delimited_list(field_name, field_info, field_value)
                    inner_type_args = get_inner_type(field_info.annotation)
                    if len(inner_type_args) != 2:
                        raise SyntaxError(
                            f"{full_name(parents, field_name)} has type {field_info.annotation} "
                            f"without exactly 2 inner types {inner_type_args}"
                        )
                    elif len(inner_type_args) == 2:
                        inner_type1 = inner_type_args[0]
                        if inner_type1 != str:
                            raise SyntaxError(
                                f"{full_name(parents, field_name)} has type {field_info.annotation} "
                                f"with the first inner type not being str (it is {inner_type1})"
                            )
                        inner_type2 = inner_type_args[1]
                        if has_sub_fields(inner_type2):
                            ref_object_dict = build_referenced_objects(
                                field_name=field_name,
                                field_info=field_info,
                                parents=parents,
                                parent_container=parent_container,
                                root_config_data=root_config_data,
                                list_of_sections=list_of_sections,
                                inner_type=inner_type2,
                            )
                            field_value = ref_object_dict
                except ValueError as e2:
                    raise ConfigError(
                        f"Field {full_name(parents, field_name)}"
                        f"Tried as list of section references and got error {e2}.\n"
                        f"Also tried as literal and got {e}"
                    )
    elif lenient_issubclass(field_info.annotation, set):
        if isinstance(field_value, str):
            field_value = set(
                parse_delimited_list(
                    field_name=field_name,
                    field_info=field_info,
                    field_value=field_value
                )
            )
    elif lenient_issubclass(field_info.annotation, frozenset):
        if isinstance(field_value, str):
            field_value = frozenset(
                parse_delimited_list(
                    field_name=field_name,
                    field_info=field_info,
                    field_value=field_value
                )
            )

    # In all cases not explicitly matched above, we'll let pydantic parse it as is
    return field_value


def match_config_data_to_model(
        model: BaseModel,
        config_data: MutableMapping,
        root_config_data: MutableMapping = None,
        parents=None
):
    if parents is None:
        parents = []
    if root_config_data is None:
        root_config_data = config_data

    # Make mappings from lower case names to actual config field names
    config_name_map = {key.lower(): key for key in config_data}
    root_config_name_map = {key.lower(): key for key in root_config_data}

    # Scan all model fields to look for values in the MutableMapping
    for field_name_outer, field_info in model.model_fields.items():
        # Check for either a direct name match or a case in-sensitive match
        field_name = field_info.alias or field_name_outer
        field_lower = field_name.lower()

        if field_lower in config_name_map:
            found = True
            source_field_name = config_name_map[field_lower]
            if source_field_name != field_name:
                config_data[field_name] = config_data[source_field_name]
                del config_data[source_field_name]
        else:
            found = False

        # Check for nested objects set using top level dotted names (e.g. [parent.child])
        # (either a direct name match or a case in-sensitive match)
        if not found and len(parents) > 0:
            section_name = '.'.join(parents + [field_name])
            section_name_lower = section_name.lower()
            if section_name in root_config_data:
                found = True
                # Copy data into place where pydantic will expect it
                config_data[field_name] = root_config_data[section_name]
            elif section_name_lower in root_config_name_map:
                found = True
                section_name = root_config_name_map[section_name_lower]
                # Copy data into place where pydantic will expect it
                config_data[field_name] = root_config_data[section_name]

        if found:
            updated_value = match_config_data_to_field_or_submodel(
                field_name=field_name,
                field_info=field_info,
                parent_container=config_data,
                root_config_data=root_config_data,
                parents=parents + [field_name]
            )
            config_data[field_name] = updated_value
    return config_data


def match_config_data_to_field_or_submodel(
        field_name: str,
        field_info: FieldInfo,
        parent_container: MutableMapping,
        root_config_data: MutableMapping = None,
        parents=None
):

    if has_sub_fields(field_info.annotation):
        if field_name not in parent_container:
            raise ValueError(f"Field {full_name(parents, field_name)} not found.")

        if isinstance(field_info, DynamicFieldInfo):
            ref_object_dict = build_referenced_objects(
                field_name=field_name,
                field_info=field_info,
                parents=parents,
                parent_container=parent_container,
                root_config_data=root_config_data,
                list_of_sections=[parent_container[field_name]],
                inner_type=field_info.annotation,
            )
            if lenient_issubclass(field_info.annotation, BaseModel):
                updated_value = list(ref_object_dict.values())[0]
            elif lenient_issubclass(field_info.annotation, list):
                updated_value = list(ref_object_dict.values())
            elif lenient_issubclass(field_info.annotation, set):
                updated_value = set(ref_object_dict.values())
            elif lenient_issubclass(field_info.annotation, frozenset):
                updated_value = frozenset(ref_object_dict.values())
            elif lenient_issubclass(field_info.annotation, dict):
                updated_value = ref_object_dict
            else:
                raise ValueError(
                    f"Field {full_name(parents, field_name)} "
                    f"DynamicFieldInfo type {field_info.annotation} not expected"
                )
        else:
            # noinspection PyTypeChecker
            updated_value = match_config_data_to_model(
                model=field_info.annotation,
                config_data=parent_container[field_name],
                root_config_data=root_config_data,
                parents=parents
            )
    else:
        updated_value = match_config_data_to_field(
            field_name=field_name,
            field_info=field_info,
            field_value=parent_container[field_name],
            parent_container=parent_container,
            root_config_data=root_config_data,
            parents=parents
        )
    return updated_value


def walk_model(
        model: BaseModel,
        parents=None
):
    if parents is None:
        parents = []

    # Scan all model fields to look for values in the MutableMapping
    for field_name, field_info in model.model_fields.items():
        if has_sub_fields(field_info.annotation):
            # noinspection PyTypeChecker
            yield from walk_model(
                model=field_info.annotation,
                parents=parents + [field_name]
            )
        else:
            yield field_name, field_info, parents



from __future__ import annotations

from typing import TypeVar, List, Any, Callable
try:
    from typing import Unpack
except ImportError:
    Unpack = 'Unpack'


from pydantic import field_validator
# noinspection PyProtectedMember
from pydantic.fields import AliasPath, AliasChoices, _FromFieldInfoInputs
from pydantic_core import PydanticUndefined

from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy
from config_wrangler.config_types.delimited_field import DelimitedListFieldInfo
from config_wrangler.validate_config_hierarchy import config_hierarchy_validator

RefConfigHierarchy = TypeVar('RefConfigHierarchy', bound=ConfigHierarchy)


class DynamicallyReferenced(ConfigHierarchy):
    """
    Represents a reference to a statically defined section of the config.
    The data type of the section can be any subclass of ConfigHierarchy.
    The validator will check that the reference exists.
    """
    ref: str

    # Note the order of decorators matters!
    # noinspection PyNestedDecorators
    @field_validator('ref')
    @classmethod
    def _validate_phase_1(cls, value):
        if value == '':
            raise ValueError('Blank is not valid for a DynamicallyReferenced section')
        return value

    @config_hierarchy_validator
    def _validate_config_hierarchy_validator(self):
        _ = self.get_referenced()

    def get_referenced(self) -> ConfigHierarchy:
        if self.ref is None:
            raise ValueError(f"DynamicallyReferenced {self} is not set")
        elif self._root_config is None:
            raise RuntimeError(
                "DynamicallyReferenced._root_config is none. "
                " Either the root class is not ConfigFromLoaders or "
                "ConfigFromLoaders.fill_hierarchy did not work."
            )
        else:
            parts = self.ref.split('.')

            model = self._root_config
            for part in parts:
                try:
                    model = getattr(model, part)
                except AttributeError:
                    raise ValueError(f"Referenced section {self.ref} not found in model {model}.")
            return model

    def __str__(self):
        return f"{self.ref}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.ref}) where get_referenced returns {self.get_referenced().__class__.__name__} instance"


class ListDynamicallyReferenced(ConfigHierarchy):
    refs: List[DynamicallyReferenced]


class DynamicFieldInfo(DelimitedListFieldInfo):
    def __init__(self, delimiter=',', **kwargs) -> None:
        super().__init__(delimiter=delimiter, **kwargs)

    @staticmethod
    def from_field(default: Any = PydanticUndefined, **kwargs: Unpack[_FromFieldInfoInputs]) -> DynamicFieldInfo:
        """Create a new `FieldInfo` object with the `Field` function.

        Args:
            default: The default value for the field. Defaults to Undefined.
            **kwargs: Additional arguments dictionary.

        Raises:
            TypeError: If 'annotation' is passed as a keyword argument.

        Returns:
            A new FieldInfo object with the given parameters.

        Example:
            This is how you can create a field with default value like this:

            ```python
            import pydantic

            class MyModel(pydantic.BaseModel):
                foo: int = pydantic.Field(4)
            ```
        """
        if 'annotation' in kwargs:
            raise TypeError('"annotation" is not permitted as a DynamicFieldInfo keyword argument')
        return DynamicFieldInfo(default=default, **kwargs)


# noinspection PyPep8Naming
def DynamicField(  # noqa: C901
        default: Any = PydanticUndefined,
        *,
        default_factory: Callable[[], Any] | None = PydanticUndefined,
        alias: str | None = PydanticUndefined,
        alias_priority: int | None = PydanticUndefined,
        validation_alias: str | AliasPath | AliasChoices | None = PydanticUndefined,
        serialization_alias: str | None = PydanticUndefined,
        title: str | None = PydanticUndefined,
        description: str | None = PydanticUndefined,
        examples: list[Any] | None = PydanticUndefined,
        exclude: bool | None = PydanticUndefined,
        include: bool | None = PydanticUndefined,
        discriminator: str | None = PydanticUndefined,
        json_schema_extra: dict[str, Any] | Callable[[dict[str, Any]], None] | None = PydanticUndefined,
        frozen: bool | None = PydanticUndefined,
        validate_default: bool | None = PydanticUndefined,
        repr: bool = PydanticUndefined,
        init_var: bool | None = PydanticUndefined,
        kw_only: bool | None = PydanticUndefined,
        pattern: str | None = PydanticUndefined,
        strict: bool | None = PydanticUndefined,
        gt: float | None = PydanticUndefined,
        ge: float | None = PydanticUndefined,
        lt: float | None = PydanticUndefined,
        le: float | None = PydanticUndefined,
        multiple_of: float | None = PydanticUndefined,
        allow_inf_nan: bool | None = PydanticUndefined,
        max_digits: int | None = PydanticUndefined,
        decimal_places: int | None = PydanticUndefined,
        min_length: int | None = PydanticUndefined,
        max_length: int | None = PydanticUndefined,
        delimiter: str = ','
) -> Any:
    """
    Create a field for a `list` of objects, plus other Pydantic `Field` configuration options.

    *Pydantic standard docs*:

    Used to provide extra information about a field, either for the model schema or complex validation. Some arguments
    apply only to number fields (`int`, `float`, `Decimal`) and some apply only to `str`.

    Args:
        default: Default value if the field is not set.
        default_factory: A callable to generate the default value, such as :func:`~datetime.utcnow`.
        alias: An alternative name for the attribute.
        alias_priority: Priority of the alias. This affects whether an alias generator is used.
        validation_alias: 'Whitelist' validation step. The field will be the single one allowed by the alias or set of
            aliases defined.
        serialization_alias: 'Blacklist' validation step. The vanilla field will be the single one of the alias' or set
            of aliases' fields and all the other fields will be ignored at serialization time.
        title: Human-readable title.
        description: Human-readable description.
        examples: Example values for this field.
        exclude: Whether to exclude the field from the model schema.
        include: Whether to include the field in the model schema.
        discriminator: Field name for discriminating the type in a tagged union.
        json_schema_extra: Any additional JSON schema data for the schema property.
        frozen: Whether the field is frozen.
        validate_default: Run validation that isn't only checking existence of defaults. `True` by default.
        repr: A boolean indicating whether to include the field in the `__repr__` output.
        init_var: Whether the field should be included in the constructor of the dataclass.
        kw_only: Whether the field should be a keyword-only argument in the constructor of the dataclass.
        strict: If `True`, strict validation is applied to the field.
            See [Strict Mode](../usage/strict_mode.md) for details.
        gt: Greater than. If set, value must be greater than this. Only applicable to numbers.
        ge: Greater than or equal. If set, value must be greater than or equal to this. Only applicable to numbers.
        lt: Less than. If set, value must be less than this. Only applicable to numbers.
        le: Less than or equal. If set, value must be less than or equal to this. Only applicable to numbers.
        multiple_of: Value must be a multiple of this. Only applicable to numbers.
        min_length: Minimum length for strings.
        max_length: Maximum length for strings.
        pattern: Pattern for strings.
        allow_inf_nan: Allow `inf`, `-inf`, `nan`. Only applicable to numbers.
        max_digits: Maximum number of allow digits for strings.
        decimal_places: Maximum number of decimal places allowed for numbers.
        delimiter: delimiter to use when parsing the input value

    Returns:
        A new [`FieldInfo`][pydantic.fields.FieldInfo], the return annotation is `Any` so `Field` can be used on
            type annotated fields without causing a typing error.
    """

    return DynamicFieldInfo.from_field(
        default,
        default_factory=default_factory,
        alias=alias,
        alias_priority=alias_priority,
        validation_alias=validation_alias,
        serialization_alias=serialization_alias,
        title=title,
        description=description,
        examples=examples,
        exclude=exclude,
        include=include,
        discriminator=discriminator,
        json_schema_extra=json_schema_extra,
        frozen=frozen,
        pattern=pattern,
        validate_default=validate_default,
        repr=repr,
        init_var=init_var,
        kw_only=kw_only,
        strict=strict,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        min_length=min_length,
        max_length=max_length,
        allow_inf_nan=allow_inf_nan,
        max_digits=max_digits,
        decimal_places=decimal_places,
        delimiter=delimiter,
    )



from typing import *

from auto_all import public
from pydantic import PrivateAttr

from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy
from config_wrangler.config_templates.password_source import PasswordSource, PasswordSourceValidated
from config_wrangler.config_types.path_types import PathFindUpExpandUser


@public
class KeepassConfig(ConfigHierarchy):
    """
    Read passwords from a keepass database.
    Performance Note:
        The default Key function AES-KDF is very slow in Python.
        Argon2d is better with Python and still secure.
        https://keepass.info/help/base/security.html#secdictprotect

    Note: A password is still needed to open & decrypt the keepass database.
    The default source of that password is KEYRING. However, any valid `PasswordSource`
    can be used. This allows
    """

    user_id: str = 'KEEPASS'
    database_path: PathFindUpExpandUser
    default_group: Optional[str] = None
    password_source: PasswordSourceValidated = PasswordSource.KEYRING
    raw_password: Optional[str] = None
    keyring_section: Optional[str] = None
    keyring_user_id: Optional[str] = None
    alternate_group_names: Dict[str, str] = {}

    _db = PrivateAttr(default=None)
    _alternate_group_names_lower = PrivateAttr(default=None)
    _keepass_credentials = PrivateAttr(default=None)

    def open_database(self, force_refresh: bool = False) -> 'pykeepass.PyKeePass':
        if self._db is None or force_refresh:
            from pykeepass import PyKeePass

            credentials_args = dict(**self.__dict__)
            if self.password_source == PasswordSource.KEYRING:
                if self.keyring_user_id is None:
                    raise ValueError(f"Keepass set to get password from KEYRING but keyring_user_id is not set")
                if self.keyring_section is None:
                    raise ValueError(f"Keepass set to get password from KEYRING but keyring_section is not set")
                credentials_args['user_id'] = self.keyring_user_id
            else:
                credentials_args['user_id'] = 'KEEPASS'

            from config_wrangler.config_templates.credentials import Credentials
            self._keepass_credentials = Credentials(**credentials_args)
            self.add_child('_keepass_credentials', self._keepass_credentials)
            keepass_encryption_password = self._keepass_credentials.get_password()
            try:
                self._db = PyKeePass(self.database_path, password=keepass_encryption_password)
            except Exception as e:
                raise ValueError(f"PyKeePass error from {self.database_path}: {repr(e)}")

            self._alternate_group_names_lower = {k.lower(): v.lower() for k, v in self.alternate_group_names.items()}
        return self._db

    def get_group_contents(self, group: str) -> Iterable['pykeepass.entry.Entry']:
        kp = self.open_database()
        results = list()
        for group_object in kp.groups:
            if group_object.name is not None and group_object.name.lower() == group.lower():
                for entry in group_object.entries:
                    results.append(entry)
        return results

    def get_group_list_contents(self, group_list: Iterable[str]) -> Iterable['pykeepass.entry.Entry']:
        results = list()
        # Make a unique set of lower case group names
        group_set = {g.lower() for g in group_list}
        for group in group_set:
            results.extend(self.get_group_contents(group))
        return results

    @staticmethod
    def _get_entry_str_value(entry_value: Any) -> str:
        if entry_value is None:
            return ''
        else:
            return str(entry_value)

    def get_password(self, group: Optional[str], title: Optional[str], user_id: Optional[str]):
        if group is None:
            if self.default_group is not None:
                group = self.default_group
            else:
                raise ValueError(f"keepass_group not provided and keepass:default_group also not provided")
        group_lower = group.lower()
        group_search_set = {group_lower}
        if self._alternate_group_names_lower is not None:
            alt_group_lower = self._alternate_group_names_lower.get(group_lower)
            if alt_group_lower is not None:
                group_search_set.add(alt_group_lower)

        selectors = [f"group.lower in {group_search_set}"]
        if title is not None:
            selectors.append(f"Title.lower = '{title.lower()}'")
            title_lower = title.lower()
        else:
            title_lower = None

        if user_id is not None:
            selectors.append(f"User_name.lower = '{user_id.lower()}'")
            user_id_lower = user_id.lower()
        else:
            user_id_lower = None

        selectors_str = ' and '.join(selectors)
        entry_matches = list()
        for entry in self.get_group_list_contents(group_search_set):
            match = True
            if title is not None:
                if self._get_entry_str_value(entry.title).lower() != title_lower:
                    match = False
            if user_id is not None:
                if self._get_entry_str_value(entry.username).lower() != user_id_lower:
                    match = False
            if match:
                entry_matches.append(entry)
        if len(entry_matches) == 0:
            raise ValueError(f"Keepass does not have entry for {selectors_str} in {self.database_path}")
        elif len(entry_matches) > 1:
            selectors_str = ' and '.join(selectors)
            matches_description_parts = list()
            for entry in entry_matches:
                matches_description_parts.append(f"<group='{entry.group}' title='{entry.title}'/>")
            raise ValueError(
                f"Keepass has multiple matches for entry for {selectors_str} in {self.database_path}."
                f"{', '.join(matches_description_parts)}"
            )
        else:
            entry_password = entry_matches[0].password
            return entry_password



import logging
import timeit
import warnings
from typing import *

from pydantic import PrivateAttr

from config_wrangler.config_templates.aws.aws_session import AWS_Session

if TYPE_CHECKING:
    # https://youtype.github.io/boto3_stubs_docs/mypy_boto3_dynamodb/
    from mypy_boto3_dynamodb.service_resource import DynamoDBServiceResource
    from mypy_boto3_dynamodb.client import DynamoDBClient
    from mypy_boto3_dynamodb.service_resource import Table
    from mypy_boto3_dynamodb.type_defs import PutItemOutputTableTypeDef
    import pynamodb.models


class DynamoDB(AWS_Session):
    _service: str = PrivateAttr(default='dynamodb')
    scan_progress_seconds: int = 10

    @property
    def resource(self) -> 'DynamoDBServiceResource':
        return super().resource

    @property
    def client(self) -> 'DynamoDBClient':
        return super().client

    def get_dynamo_table(
            self,
            dynamo_table_name,
            region_name: str = None
    ) -> 'Table':
        if region_name is not None:
            warnings.warn(
                "DynamoDB.get_dynamo_table argument region_name is deprecated. "
                "Use DynamoDB.region_name property instead."
            )
            if self.region_name != region_name:
                warnings.warn(
                    f"DynamoDB.get_dynamo_table argument region_name value '{region_name}' conflicts with "
                    f"DynamoDB.region_name value '{self.region_name}! "
                    "Deprecated argument value was ignored."
                )
        dynamodb = self.resource

        return dynamodb.Table(dynamo_table_name)

    def query_dynamo_table(
            self,
            dynamo_table: 'Table',
            scan_args_list: Iterable[dict],
    ) -> Iterable[dict]:
        log = logging.getLogger('DynamoDB')
        start_time = timeit.default_timer()
        page = 1
        for scan_args in scan_args_list:
            page += 1
            tbl_data = dynamo_table.query(**scan_args)
            for item in tbl_data['Items']:
                yield item

            while 'LastEvaluatedKey' in tbl_data:
                page += 1
                if (timeit.default_timer() - start_time) > self.scan_progress_seconds:
                    log.info(f"Processing query page {page:,}")
                    start_time = timeit.default_timer()
                tbl_data = dynamo_table.query(ExclusiveStartKey=tbl_data['LastEvaluatedKey'], **scan_args)
                for item in tbl_data['Items']:
                    yield item

    def query_dynamo_table_by_name(
            self,
            dynamo_table_name: str,
            scan_args_list: Iterable[dict],
            region_name: str = None,  # Deprecated
    ) -> Iterable[dict]:
        dynamo_table = self.get_dynamo_table(dynamo_table_name, region_name=region_name)
        yield from self.query_dynamo_table(dynamo_table, scan_args_list)

    def scan_dynamo_table(
        self,
        dynamo_table: 'Table',
    ) -> Iterator[dict]:
        log = logging.getLogger('DynamoDB')
        start_time = timeit.default_timer()
        tbl_data = dynamo_table.scan()
        for item in tbl_data['Items']:
            yield item

        page = 1
        while 'LastEvaluatedKey' in tbl_data:
            page += 1
            if (timeit.default_timer() - start_time) > self.scan_progress_seconds:
                log.info(f"Processing scan page {page:,}")
                start_time = timeit.default_timer()
            tbl_data = dynamo_table.scan(ExclusiveStartKey=tbl_data['LastEvaluatedKey'])
            for item in tbl_data['Items']:
                yield item

    def scan_dynamo_table_by_name(
            self,
            dynamo_table_name: str,
            region_name: str = None  # Deprecated
    ) -> Iterable[dict]:
        dynamo_table = self.get_dynamo_table(dynamo_table_name, region_name=region_name)
        yield from self.scan_dynamo_table(dynamo_table)


class DynamoDBTable(DynamoDB):
    table_name: str

    def get_dynamo_table(self, **kwargs) -> 'Table':
        parent_table_arg = 'dynamo_table_name'
        if parent_table_arg in kwargs and kwargs[parent_table_arg] is not None:
            return super().get_dynamo_table(dynamo_table_name=kwargs[parent_table_arg])
        else:
            return super().get_dynamo_table(dynamo_table_name=self.table_name)

    def query_dynamo_table(
            self,
            scan_args_list: Iterable[dict],
            **kwargs
    ) -> Iterable[dict]:
        parent_table_arg = 'dynamo_table_name'
        if parent_table_arg in kwargs and kwargs[parent_table_arg] is not None:
            return super().query_dynamo_table(
                dynamo_table=kwargs[parent_table_arg],
                scan_args_list=scan_args_list,
            )
        else:
            return super().query_dynamo_table(
                dynamo_table=self.get_dynamo_table(),
                scan_args_list=scan_args_list,
            )

    def scan_dynamo_table(
        self,
        **kwargs
    ):
        parent_table_arg = 'dynamo_table'
        if parent_table_arg in kwargs and kwargs[parent_table_arg] is not None:
            return super().scan_dynamo_table(
                dynamo_table=kwargs[parent_table_arg],
            )
        else:
            return super().scan_dynamo_table(
                dynamo_table=self.get_dynamo_table(),
            )

    def put_item(self, item: Mapping[str, Any]) -> 'PutItemOutputTableTypeDef':
        return self.get_dynamo_table().put_item(Item=item)

    def get_connected_pynamodb(
        self,
        model: 'pynamodb.models.Model',
        connect_table_name: Optional[str] = None
    ) -> 'pynamodb.models.Model':
        # Optional library
        from pynamodb.indexes import Index

        if connect_table_name is None:
            connect_table_name = self.table_name or model.Meta.table_name

        # # We need instance specific versions of the Model class
        class _ConnectedClass(model):
            class Meta:
                connected = True
                table_name = connect_table_name

        _ConnectedClass.Meta.region = self.region_name
        _ConnectedClass.Meta.aws_access_key_id = self.user_id
        _ConnectedClass.Meta.aws_secret_access_key = self.get_password()
        # Optional, only for temporary credentials like those received when assuming a role
        credentials = self.session.get_credentials()
        _ConnectedClass.Meta.aws_session_token = credentials.token

        for attribute_name in dir(_ConnectedClass):
            attribute = getattr(_ConnectedClass, attribute_name)
            if isinstance(attribute, Index):

                class InnerIndex(attribute.__class__):
                    class Meta:
                        index_name = attribute.Meta.index_name
                        projection = attribute.Meta.projection

                index_obj = InnerIndex()
                setattr(_ConnectedClass, attribute_name, index_obj)
                index_obj.Meta.model = _ConnectedClass
                index_obj._model = _ConnectedClass
        return _ConnectedClass



import logging
from pathlib import Path
from typing import *

from config_wrangler.config_data_loaders.env_config_data_loader import EnvConfigDataLoader
from config_wrangler.config_data_loaders.ini_config_data_loader import IniConfigDataLoader
from config_wrangler.config_from_loaders import ConfigFromLoaders


class ConfigFromIniEnv(ConfigFromLoaders):
    # noinspection PyMethodParameters
    def __init__(
            __pydantic_self__,
            file_name: str = 'config.ini',
            start_path: Optional[Union[str, Path]] = None,
            config_load_log_level: int = logging.INFO,
            **kwargs: Dict[str, Any]
    ) -> None:
        env_loader = EnvConfigDataLoader()
        ini_loader = IniConfigDataLoader(start_path=start_path, file_name=file_name)
        super().__init__(
            _config_data_loaders=[env_loader, ini_loader],
            config_load_log_level=config_load_log_level,
            **kwargs
        )
